from .task_execution import *
from .train import ModelTrainer
from .data import *


# maybe some functionality for Runner to rerun if necessary but not print message that they are already exist if they do
# add more bait configurations


class RunConfigList:
    """
    Helper class holding a list of experiment configurations to be executed.
    """
    def __init__(self, configs: Optional[List[Tuple[float, ModelTrainer, float]]] = None):
        """
        Constructor.
        :param configs: Initialize with configs or empty list if configs is None.
        """
        self.configs = configs or []

    def append(self, ram_gb_per_sample: float, trainer: ModelTrainer, ram_gb_per_sample_bs: float = 0.0):
        """
        Append a configuration
        :param ram_gb_per_sample: RAM GB per sample constant used for RAM usage estimation of this configuration
        :param trainer: ModelTrainer that can be used to run BMDAL and contains the BMDAL configuration
        :param ram_gb_per_sample_bs: RAM GB per (sample * batch size) used for RAM usage estimation.
        This can be used for an additional term on top of ram_gb_per_sample, but is only helpful for some methods
        whose RAM usage increases significantly with the batch size.
        """
        self.configs.append((ram_gb_per_sample, trainer, ram_gb_per_sample_bs))

    def filter_names(self, alg_names: List[str]):
        """
        Filter out all configurations whose alg_names are not contained in alg_names.
        :param alg_names:
        :return:
        """
        return RunConfigList([(ram_gb_per_sample, trainer, ram_gb_per_sample_float)
                              for ram_gb_per_sample, trainer, ram_gb_per_sample_float in self.configs
                              if trainer.alg_name in alg_names])

    def __iter__(self) -> Iterable[Tuple[float, ModelTrainer, float]]:
        """
        Allows to iterate over the configs, e.g. within a for loop:
        for config in self:
            print(config)
        :return: Returns an iterator that outputs tuples (ram_gb_per_sample, trainer, ram_gb_per_sample_bs).
        """
        for cfg in self.configs:
            yield cfg


def get_bmdal_configs(**kwargs) -> RunConfigList:
    """
    :param kwargs: allows to set some hyperparameters, for example the learning rate, sigma_w, sigma_b, etc.
    :return: Returns a list of configurations for BMDAL used in the paper.
    """
    sigma = kwargs.get('post_sigma', 0.1)
    compute_eff_dim = True
    kwargs = utils.update_dict(dict(maxdet_sigma=sigma, bait_sigma=sigma, compute_eff_dim=compute_eff_dim,
                                    allow_float64=True, lr=0.375, weight_gain=0.2, bias_gain=0.2), kwargs)
    
    lst = RunConfigList()

    lst.append(1e-6, ModelTrainer(f'NN_random', selection_method='random',
                                         base_kernel='linear', kernel_transforms=[], **kwargs))

    # bait kernel comparison
    for fb_mode, overselection_factor in [('f', 1.0), ('fb', 2.0)]:
        lst.append(2e-5, ModelTrainer(f'NN_bait-{fb_mode}-p_ll_train', selection_method='bait',
                                      overselection_factor=overselection_factor, base_kernel='ll',
                                      kernel_transforms=[('train', [sigma, None])],
                                      **kwargs))
        lst.append(2e-5, ModelTrainer(f'NN_bait-{fb_mode}-p_grad_rp-512_train', selection_method='bait',
                                      overselection_factor=overselection_factor, base_kernel='grad',
                                      kernel_transforms=[('rp', [512]), ('train', [sigma, None])],
                                      **kwargs))
        if fb_mode == 'f':
            lst.append(2e-5, ModelTrainer(f'NN_bait-{fb_mode}-p_linear_train', selection_method='bait',
                                          overselection_factor=overselection_factor,
                                          base_kernel='linear', kernel_transforms=[('train', [sigma, None])],
                                          **kwargs))
            lst.append(2e-5, ModelTrainer(f'NN_bait-{fb_mode}-p_grad_ens-3_rp-512_train', selection_method='bait',
                                          overselection_factor=overselection_factor, base_kernel='grad', n_models=3,
                                          kernel_transforms=[('ens', []), ('rp', [512]),
                                                             ('train', [sigma, None])],
                                          **kwargs))
            lst.append(2e-5, ModelTrainer(f'NN_bait-{fb_mode}-p_ll_ens-3_rp-512_train', selection_method='bait',
                                          overselection_factor=overselection_factor, n_models=3,
                                          base_kernel='ll', kernel_transforms=[('ens', []),
                                                                               ('rp', [512]),
                                                                               ('train', [sigma, None])],
                                          **kwargs))

    # maxdet kernel comparison
    lst.append(2e-5, ModelTrainer(f'NN_maxdet-p_ll_train', selection_method='maxdet',
                                         base_kernel='ll', kernel_transforms=[('train', [sigma, None])],
                                         **kwargs), 8e-9)
    lst.append(2.5e-5, ModelTrainer(f'NN_maxdet-p_grad_rp-512_train', selection_method='maxdet',
                                         base_kernel='grad',
                                         kernel_transforms=[('rp', [512]), ('train', [sigma, None])],
                                         **kwargs), 8e-9)
    lst.append(2.5e-5, ModelTrainer(f'NN_maxdet-p_grad_ens-3_rp-512_train', selection_method='maxdet',
                                         base_kernel='grad', n_models=3,
                                         kernel_transforms=[('ens', []), ('rp', [512]),
                                                            ('train', [sigma, None])],
                                         **kwargs), 8e-9)
    lst.append(2.5e-5,
               ModelTrainer(f'NN_maxdet-p_grad_rp-512_acs-rf-512', selection_method='maxdet',
                            base_kernel='grad',
                            kernel_transforms=[('rp', [512]), ('acs-rf', [512, sigma, None])],
                            **kwargs), 8e-9)
    lst.append(2.5e-5,
               ModelTrainer(f'NN_maxdet-p_grad_rp-512_acs-grad', selection_method='maxdet',
                            base_kernel='grad',
                            kernel_transforms=[('rp', [512]), ('acs-grad', [sigma, None])],
                            **kwargs), 8e-9)
    lst.append(2.5e-5,
               ModelTrainer(f'NN_maxdet-p_grad_rp-512_acs-rf-hyper-512', selection_method='maxdet',
                            base_kernel='grad',
                            kernel_transforms=[('rp', [512]), ('acs-rf-hyper', [512, None])],
                            **kwargs), 8e-9)
    lst.append(2e-5,
               ModelTrainer(f'NN_maxdet-p_ll_acs-rf-512', selection_method='maxdet',
                            base_kernel='ll',
                            kernel_transforms=[('acs-rf', [512, sigma, None])],
                            **kwargs), 8e-9)

    lst.append(8e-5, ModelTrainer(f'NN_maxdet-tp_grad_scale', selection_method='maxdet',
                                         base_kernel='grad', sel_with_train=True,
                                         kernel_transforms=[('scale', [None])],
                                         **kwargs), 8e-9)
    lst.append(2e-5, ModelTrainer(f'NN_maxdet-p_ll_ens-3_rp-512_train', selection_method='maxdet',
                                         n_models=3,
                                         base_kernel='ll', kernel_transforms=[('ens', []),
                                                                              ('rp', [512]),
                                                                              ('train', [sigma, None])],
                                         **kwargs), 8e-9)
    lst.append(8e-5, ModelTrainer(f'NN_maxdet-tp_nngp_scale', selection_method='maxdet',
                                         base_kernel='nngp', sel_with_train=True,
                                         kernel_transforms=[('scale', [None])],
                                         **kwargs), 8e-9)
    lst.append(2e-5, ModelTrainer(f'NN_maxdet-p_linear_train', selection_method='maxdet',
                                         base_kernel='linear', kernel_transforms=[('train', [sigma, None])],
                                         **kwargs), 8e-9)

    # maxdist, kmeanspp, lcmd kernel comparisons
    for sel_name in ['maxdist', 'kmeanspp', 'lcmd']:
        lst.append(2e-6, ModelTrainer(f'NN_{sel_name}-tp_linear', selection_method=sel_name,
                                             base_kernel='linear', kernel_transforms=[], **kwargs))
        lst.append(2e-6, ModelTrainer(f'NN_{sel_name}-tp_nngp', selection_method=sel_name,
                                             base_kernel='nngp', kernel_transforms=[], **kwargs))
        lst.append(4e-6, ModelTrainer(f'NN_{sel_name}-tp_ll', selection_method=sel_name,
                                             base_kernel='ll', kernel_transforms=[], **kwargs))
        lst.append(2e-5, ModelTrainer(f'NN_{sel_name}-tp_grad', selection_method=sel_name,
                                             base_kernel='grad', kernel_transforms=[], **kwargs))
        lst.append(4e-6, ModelTrainer(f'NN_{sel_name}-tp_grad_rp-512', selection_method=sel_name,
                                             base_kernel='grad', kernel_transforms=[('rp', [512])], **kwargs))
        lst.append(8e-6, ModelTrainer(f'NN_{sel_name}-tp_grad_ens-3_rp-512', selection_method=sel_name,
                                             n_models=3,
                                             base_kernel='grad', kernel_transforms=[('ens', []), ('rp', [512])],
                                             **kwargs))
        lst.append(2e-5, ModelTrainer(f'NN_{sel_name}-p_ll_train', selection_method=sel_name,
                                             sel_with_train=False,
                                             base_kernel='ll', kernel_transforms=[('train', [sigma, None])],
                                             **kwargs))
        lst.append(2e-5, ModelTrainer(f'NN_{sel_name}-p_grad_rp-512_train', selection_method=sel_name,
                                             sel_with_train=False,
                                             base_kernel='grad', kernel_transforms=[('rp', [512]),
                                                                                    ('train', [sigma, None])],
                                             **kwargs))
        lst.append(1e-5, ModelTrainer(f'NN_{sel_name}-tp_ll_ens-3_rp-512', selection_method=sel_name,
                                             n_models=3,
                                             base_kernel='ll', kernel_transforms=[('ens', []), ('rp', [512])],
                                             **kwargs))
        lst.append(2e-5,
                   ModelTrainer(f'NN_{sel_name}-p_grad_rp-512_acs-rf-512', selection_method=sel_name,
                                sel_with_train=False,
                                base_kernel='grad', kernel_transforms=[('rp', [512]),
                                                                       ('acs-rf', [512, sigma, None])],
                                **kwargs))
        lst.append(2e-5,
                   ModelTrainer(f'NN_{sel_name}-p_grad_rp-512_acs-grad', selection_method=sel_name,
                                sel_with_train=False,
                                base_kernel='grad', kernel_transforms=[('rp', [512]),
                                                                       ('acs-grad', [sigma, None])],
                                **kwargs))
        lst.append(2e-5,
                   ModelTrainer(f'NN_{sel_name}-p_grad_rp-512_acs-rf-hyper-512', selection_method=sel_name,
                                sel_with_train=False,
                                base_kernel='grad', kernel_transforms=[('rp', [512]),
                                                                       ('acs-rf-hyper', [512, None])],
                                **kwargs))
        lst.append(2e-5, ModelTrainer(f'NN_{sel_name}-p_ll_acs-rf-512', selection_method=sel_name,
                                             sel_with_train=False,
                                             base_kernel='ll',
                                             kernel_transforms=[('acs-rf', [512, sigma, None])],
                                             **kwargs))

    # maxdiag kernel comparison
    lst.append(8e-6, ModelTrainer(f'NN_maxdiag_ll_train', selection_method='maxdiag',
                                         base_kernel='ll', kernel_transforms=[('train', [sigma, None])],
                                         **kwargs))
    lst.append(8e-6, ModelTrainer(f'NN_maxdiag_grad_rp-512_train', selection_method='maxdiag',
                                         base_kernel='grad',
                                         kernel_transforms=[('rp', [512]), ('train', [sigma, None])],
                                         **kwargs))
    lst.append(8e-6, ModelTrainer(f'NN_maxdiag_grad_ens-3_rp-512_train', selection_method='maxdiag',
                                         base_kernel='grad', n_models=3,
                                         kernel_transforms=[('ens', []), ('rp', [512]),
                                                            ('train', [sigma, None])],
                                         **kwargs))
    lst.append(1.3e-5,
               ModelTrainer(f'NN_maxdiag_grad_rp-512_acs-rf-512', selection_method='maxdiag',
                            base_kernel='grad',
                            kernel_transforms=[('rp', [512]), ('acs-rf', [512, sigma, None])],
                            **kwargs))
    lst.append(1.3e-5,
               ModelTrainer(f'NN_maxdiag_grad_rp-512_acs-grad', selection_method='maxdiag',
                            base_kernel='grad',
                            kernel_transforms=[('rp', [512]), ('acs-grad', [sigma, None])],
                            **kwargs))
    lst.append(1.3e-5,
               ModelTrainer(f'NN_maxdiag_grad_rp-512_acs-rf-hyper-512', selection_method='maxdiag',
                            base_kernel='grad',
                            kernel_transforms=[('rp', [512]), ('acs-rf-hyper', [512, None])],
                            **kwargs))
    lst.append(1.3e-5,
               ModelTrainer(f'NN_maxdiag_ll_acs-rf-512', selection_method='maxdiag',
                            base_kernel='ll',
                            kernel_transforms=[('acs-rf', [512, sigma, None])]))
    lst.append(8e-6, ModelTrainer(f'NN_maxdiag_linear_train', selection_method='maxdiag',
                                         base_kernel='linear', kernel_transforms=[('train', [sigma, None])],
                                         **kwargs))
    lst.append(8e-5, ModelTrainer(f'NN_maxdiag_nngp_train', selection_method='maxdiag',
                                         base_kernel='nngp', kernel_transforms=[('train', [sigma, None])],
                                         **kwargs))
    lst.append(8e-5, ModelTrainer(f'NN_maxdiag_grad_train', selection_method='maxdiag',
                                         base_kernel='grad', kernel_transforms=[('train', [sigma, None])],
                                         **kwargs))

    # Frank-Wolfe kernel comparison
    lst.append(2e-5, ModelTrainer(f'NN_fw-p_ll_train', selection_method='fw',
                                         base_kernel='ll', kernel_transforms=[('train', [sigma, None])],
                                         **kwargs))
    lst.append(2e-5, ModelTrainer(f'NN_fw-p_grad_rp-512_train', selection_method='fw',
                                         base_kernel='grad', kernel_transforms=[('rp', [512]),
                                                                                ('train', [sigma, None])],
                                         **kwargs))
    lst.append(2e-5, ModelTrainer(f'NN_fw-p_ll_acs-grad_rp-512', selection_method='fw',
                                         base_kernel='ll', kernel_transforms=[('acs-grad', [sigma, None]),
                                                                              ('rp', [512])],
                                         **kwargs))
    lst.append(2e-5, ModelTrainer(f'NN_fw-p_grad_rp-512_acs-grad_rp-512', selection_method='fw',
                                         base_kernel='grad', kernel_transforms=[('rp', [512]),
                                                                                ('acs-grad', [sigma, None]),
                                                                                ('rp', [512])],
                                         **kwargs))
    lst.append(2e-5, ModelTrainer(f'NN_fw-p_ll_acs-rf-512', selection_method='fw',
                                         base_kernel='ll', kernel_transforms=[('acs-rf', [512, sigma, None])],
                                         **kwargs))

    lst.append(2e-5, ModelTrainer(f'NN_fw-p_grad_rp-512_acs-rf-512', selection_method='fw',
                                         base_kernel='grad', kernel_transforms=[('rp', [512]),
                                                                                ('acs-rf', [512, sigma, None])],
                                         **kwargs))
    lst.append(2e-5, ModelTrainer(f'NN_fw-p_ll_acs-rf-hyper-512', selection_method='fw',
                                         base_kernel='ll',
                                         kernel_transforms=[('acs-rf-hyper', [512, None])],
                                         **kwargs))
    lst.append(2e-5,
               ModelTrainer(f'NN_fw-p_grad_rp-512_acs-rf-hyper-512', selection_method='fw',
                            base_kernel='grad',
                            kernel_transforms=[('rp', [512]), ('acs-rf-hyper', [512, None])],
                            **kwargs))

    return lst


def get_relu_tuning_configs() -> RunConfigList:
    lst = RunConfigList()

    for lr in [3e-2, 5e-2, 8e-2]:
        for sigma_w in [0.25, 0.4, 0.7, 1.0, 1.414]:
            for wd in [1e-2, 1e-3, 0.0]:
                lst.append(1e-6, ModelTrainer(f'NN_lr-{lr:g}_sigmaw-{sigma_w:g}_wd-{wd:g}',
                                                     lr=lr, weight_decay=wd, weight_gain=sigma_w,
                                                     selection_method='random', base_kernel='linear',
                                                     kernel_transforms=[]))
    for lr in [8e-2, 1e-1, 2e-1]:
        for sigma_w in [0.25, 0.4, 0.5]:
            for wd in [1e-2, 1e-3, 1e-1, 0.0, 3e-3]:
                lst.append(1e-6, ModelTrainer(f'NN_lr-{lr:g}_sigmaw-{sigma_w:g}_wd-{wd:g}',
                                                     lr=lr, weight_decay=wd, weight_gain=sigma_w,
                                                     selection_method='random', base_kernel='linear',
                                                     kernel_transforms=[]))

    for lr in [2e-1, 3e-1, 4e-1]:
        for sigma_w in [0.25]:
            for wd in [0.0, 1e-3, 3e-3]:
                lst.append(1e-6, ModelTrainer(f'NN_lr-{lr:g}_sigmaw-{sigma_w:g}_wd-{wd:g}',
                                                     lr=lr, weight_decay=wd, weight_gain=sigma_w,
                                                     selection_method='random', base_kernel='linear',
                                                     kernel_transforms=[]))

    for lr in [7.5e-2]:
        for sigma_w in [1.0]:
            for wd in [0.0]:
                for wig in [0.25]:
                    for sigma_b in [0.1, 0.4, 1.0]:
                        lst.append(1e-6,
                                   ModelTrainer(
                                       f'NN_lr-{lr:g}_sigmaw-{sigma_w:g}_wd-{wd:g}_wig-{wig:g}-sigmab-{sigma_b:g}',
                                       lr=lr, weight_decay=wd, weight_gain=sigma_w,
                                       weight_init_gain=wig, bias_gain=sigma_b,
                                       selection_method='random', base_kernel='linear',
                                       kernel_transforms=[]))

    for lr in [7.5e-2]:
        for sigma_w in [1.0]:
            for wd in [0.0]:
                for wig in [0.1, 1.0]:
                    for sigma_b in [1.0]:
                        lst.append(1e-6,
                                   ModelTrainer(
                                       f'NN_lr-{lr:g}_sigmaw-{sigma_w:g}_wd-{wd:g}_wig-{wig:g}-sigmab-{sigma_b:g}',
                                       lr=lr, weight_decay=wd, weight_gain=sigma_w,
                                       weight_init_gain=wig, bias_gain=sigma_b,
                                       selection_method='random', base_kernel='linear',
                                       kernel_transforms=[]))

    for lr in [7.5e-2, 1e-1]:
        for sigma_w in [1.0]:
            for wd in [0.0]:
                for wig in [0.1, 0.25, 0.5]:
                    for sigma_b in [1.0]:
                        lst.append(1e-6,
                                   ModelTrainer(
                                       f'NN_lr-{lr:g}_sigmaw-{sigma_w:g}_wd-{wd:g}_wig-{wig:g}-sigmab-{sigma_b:g}',
                                       lr=lr, weight_decay=wd, weight_gain=sigma_w,
                                       weight_init_gain=wig, bias_gain=sigma_b,
                                       selection_method='random', base_kernel='linear',
                                       kernel_transforms=[]))

    for sigma_w in [0.1, 0.2, 0.25, 0.3]:
        for sigma_b in [0.1, 0.2, 0.3]:
            for lr in [5e-2 / sigma_w, 7.5e-2 / sigma_w, 1e-1 / sigma_w]:
                wd = 0.0
                wig = 1.0
                lst.append(1e-6,
                           ModelTrainer(f'NN_lr-{lr:g}_sigmaw-{sigma_w:g}_wd-{wd:g}_wig-{wig:g}-sigmab-{sigma_b:g}',
                                        lr=lr, weight_decay=wd, weight_gain=sigma_w,
                                        weight_init_gain=wig, bias_gain=sigma_b,
                                        selection_method='random', base_kernel='linear',
                                        kernel_transforms=[]))

    sigma_w = 0.2
    sigma_b = 0.2
    for lr in [5e-2 / sigma_w, 7.5e-2 / sigma_w, 1e-1 / sigma_w]:
        for lr_sched in ['hat', 'warmup']:
            wd = 0.0
            wig = 1.0
            lst.append(1e-6,
                       ModelTrainer(
                           f'NN_lr-{lr:g}_sigmaw-{sigma_w:g}_wd-{wd:g}_wig-{wig:g}-sigmab-{sigma_b:g}-{lr_sched}',
                           lr=lr, weight_decay=wd, weight_gain=sigma_w,
                           weight_init_gain=wig, bias_gain=sigma_b, lr_sched=lr_sched,
                           selection_method='random', base_kernel='linear',
                           kernel_transforms=[]))

    return lst


def get_silu_tuning_configs() -> RunConfigList:
    lst = RunConfigList()

    for sigma_w in [0.2, 0.5, 1.0]:
        for sigma_b in [0.1, 0.25, 0.5, 1.0]:
            for lr in [3e-2 / sigma_w, 5e-2 / sigma_w, 7.5e-2 / sigma_w, 1e-1 / sigma_w]:
                wd = 0.0
                wig = 1.0
                lst.append(1e-6,
                           ModelTrainer(f'NN_sigmaw-{sigma_w:g}_wd-{wd:g}_sigmab-{sigma_b:g}_lr-{lr:g}',
                                        lr=lr, weight_decay=wd, weight_gain=sigma_w,
                                        weight_init_gain=wig, bias_gain=sigma_b, act='silu',
                                        selection_method='random', base_kernel='linear',
                                        kernel_transforms=[]))

    for wd in [1e-2, 1e-3]:
        sigma_w = 0.5
        sigma_b = 1.0
        lr = 0.15
        wig = 1.0
        lst.append(1e-6,
                   ModelTrainer(f'NN_sigmaw-{sigma_w:g}_wd-{wd:g}_sigmab-{sigma_b:g}_lr-{lr:g}',
                                lr=lr, weight_decay=wd, weight_gain=sigma_w,
                                weight_init_gain=wig, bias_gain=sigma_b, act='silu',
                                selection_method='random', base_kernel='linear',
                                kernel_transforms=[]))

    return lst


def run_experiments(exp_name: str, n_splits: int, run_config_list: RunConfigList,
                    batch_sizes_configs: Optional[List[List[int]]] = None, task_descs: Optional[List[str]] = None,
                    use_pool_for_normalization: bool = True, max_jobs_per_device: int = 20,
                    n_train_initial: int = 256, ds_names: Optional[List[str]] = None,
                    sequential_split: Optional[int] = 9):
    """
    This function allows to run experiments in a parallelized fashion.
    :param exp_name: Name for the group of experiments. This name will be used as a folder name.
    :param run_config_list: List of configurations that should be run.
    :param n_splits: Number of random splits to run.
    :param batch_sizes_configs: Optional list of lists of batch sizes.
    The callback function will be called once for each list of batch sizes,
    with tasks using this list of batch sizes for BMDAL.
    By default, batch_sizes_configs=[[256]*16] will be used as in the paper.
    :param task_descs: Optional list of task descriptions, which will be appended to the dataset names.
    One task description per list of batch sizes in batch_sizes_configs should be provided.
    :param use_pool_for_normalization: If True, compute the statistics for standardizing the inputs of the data sets
    based on the initial training and pool set. Otherwise, compute them only on the initial training set.
    :param max_jobs_per_device: Maximum number of processes run per device.
    If GPUs are available, each GPU is one device. Otherwise, the CPU is used as a single device.
    :param n_train_initial: Initial training set size. Defaults to 256 as in the paper.
    :param ds_names: Names of data sets that should be used. By default, all data sets from the benchmark are used.
    :param sequential_split: ID of the random split where max_jobs_per_device is set to 1
    for accurate timing statistics. Defaults to 9. If no split should be used for timing, set this to None.
    """
    if ds_names is None:
        ds_names = ['sgemm', 'mlr_knn_rng', 'wecs', 'ct', 'kegg_undir_uci', 'online_video', 'query_agg_count',
                    'poker', 'road_network', 'fried', 'diamonds', 'methane', 'protein', 'sarcos', 'stock']

    if batch_sizes_configs is None:
        batch_sizes_configs = [[256] * 16]
    if task_descs is None:
        task_descs = ['256x16']

    tabular_tasks = Task.get_tabular_tasks(n_train=n_train_initial, al_batch_sizes=[], ds_names=ds_names)

    for t in tabular_tasks:
        print(f'Task {t.task_name} has n_pool={t.n_pool}, n_test={t.n_test}, n_features={t.data_info.n_features}')

    for max_split_id in range(0, n_splits):
        # run each split sequentially
        print(f'Running all configurations on split {max_split_id}')
        # run only one experiment per GPU on split sequential_split for timing experiments
        do_timing = max_split_id == sequential_split
        scheduler = JobScheduler(max_jobs_per_device=1 if do_timing else max_jobs_per_device)
        runner = JobRunner(scheduler=scheduler)
        for split_id in range(0, max_split_id+1):
            for batch_sizes_config, task_desc in zip(batch_sizes_configs, task_descs):
                tasks = Task.get_tabular_tasks(n_train=n_train_initial, al_batch_sizes=batch_sizes_config,
                                               ds_names=ds_names,
                                               desc=task_desc)
                for ram_gb_per_sample, trainer, ram_gb_per_sample_bs in run_config_list:
                    runner.add(exp_name, split_id, tasks, ram_gb_per_sample, trainer, do_timing=do_timing,
                               warn_if_exists=(split_id == max_split_id),
                               use_pool_for_normalization=use_pool_for_normalization,
                               ram_gb_per_sample_bs=ram_gb_per_sample_bs)
        runner.run_all()


def get_relu_configs() -> RunConfigList:
    return get_bmdal_configs(weight_gain=0.2, bias_gain=0.2, post_sigma=1e-3, lr=0.375, act='relu')


def get_silu_configs() -> RunConfigList:
    return get_bmdal_configs(weight_gain=0.5, bias_gain=1.0, post_sigma=1e-3, lr=0.15, act='silu')


if __name__ == '__main__':
    use_pool_for_normalization = True

    relu_bs_configs = get_relu_configs().filter_names(
        ['NN_lcmd-tp_grad_rp-512', 'NN_kmeanspp-p_grad_rp-512_acs-rf-512', 'NN_fw-p_grad_rp-512_acs-rf-hyper-512',
         'NN_maxdist-p_grad_rp-512_train',
         'NN_maxdet-p_grad_rp-512_train',
         'NN_maxdiag_grad_rp-512_acs-rf-512',
         'NN_bait-f-p_grad_rp-512_train'])

    # # ReLU experiments
    run_experiments('relu', 20, get_relu_configs(),
                    use_pool_for_normalization=use_pool_for_normalization)
    # SiLU experiments, without batch size experiments
    run_experiments('silu', 20, get_silu_configs(),
                    use_pool_for_normalization=use_pool_for_normalization)
    # # ReLU batch size experiments
    run_experiments('relu', 20, relu_bs_configs,
                    batch_sizes_configs=[[2**(12-m)]*(2**m) for m in range(7) if m != 4],
                    task_descs=[f'{2**(12-m)}x{2**m}' for m in range(7) if m != 4],
                    use_pool_for_normalization=use_pool_for_normalization)

    # for hyperparameter optimization
    # run_experiments('relu_tuning', 2, get_relu_tuning_configs(),
    #                 use_pool_for_normalization=use_pool_for_normalization)
    # run_experiments('silu_tuning', 2, get_silu_tuning_configs(),
    #                 use_pool_for_normalization=use_pool_for_normalization)



import torch
import torch.nn as nn

from ..layers import *
from .features import *
from .selection import *


def select_batch(batch_size: int, models: List[nn.Module], data: Dict[str, FeatureData],
                 y_train: Optional[torch.Tensor],
                 base_kernel: str, kernel_transforms: List[Tuple[str, List]], selection_method: str,
                 precomp_batch_size: int = 32768, nn_batch_size=8192, **config) \
        -> Tuple[torch.Tensor, Dict[str, Any]]:
    """
    This method allows to apply the methods from the paper for selecting a batch of indices from the pool set,
    where the base kernel, kernel transformations and selection method can be easily configured.
    This method may reset the gradients of the parameters of provided models to None.
    :param batch_size: Number of samples to select from the pool set.
    :param models: List of NN models that are used for certain base kernels.
    Only one model is needed except if ensembling should be used.
    :param data: A dictionary such that data['train'] contains a FeatureData object representing the training data,
    and data['pool'] contains a FeatureData object representing the pool data.
    If the base kernels 'grad' or 'll' should be used, the get_tensor() method of the FeatureData objects
    should return tensors in a format that can be fed to the specified models.
    :param y_train: Tensor representing the training labels
    (which should be in the same order as the inputs in data['train']).
    This parameter is only used for the acs-rf-hyper transformation
    and can be set to None if this transformation is not used.
    :param base_kernel: Base kernel to use. Currently supported base kernels are
    'll', 'grad', 'lin', 'nngp', 'ntk', and 'laplace'.
    :param kernel_transforms: List of kernel transformations.
    Each kernel transformation is given by a tuple (name, args)
    and the corresponding transformation method is then called with parameters *args.
    The following kernel transformations are currently supported:
    ('train', [sigma]) applies k_{\to\Xtrain}
    ('train', [sigma, factor]) applies (factor^2 * k)_{\to post(\Xtrain, sigma^2)},
                                if factor is not None, or k_{\to\Xtrain} otherwise
    ('pool', [sigma]) does the same for \Xpool instead of \Xtrain
    ('pool', [sigma, factor]) does the same for \Xpool instead of \Xtrain
    ('scale', []) applies k_{\to scale(\Xtrain)}
    ('scale', [factor]) scales to factor^2 * k, if factor is not None, and k_{\to scale(\Xtrain)} otherwise
    ('rp', [n_features]) applies sketching (= random projections) with n_features random features
    ('ens', []) applies k_{\to ens}, i.e. ensembling of all kernels.
    ('acs-rf', [n_features, sigma]) applies k_{\to acs-rf(n_features)} with GP noise standard deviation sigma.
                                    As for 'train', there is also an optional third parameter
                                    for explicitly controlling the rescaling.
    ('acs-rf-hyper', [n_features]) applies k_{\to acs-rf-hyper(n_features)}.
                                    As for 'train', there is also an optional second parameter
                                    for explicitly controlling the rescaling.
    ('acs-grad', [sigma]) applies k_{\to acs-grad} with GP noise standard deviation sigma.
                                    As for 'train', there is also an optional second parameter
                                    for explicitly controlling the rescaling.
    :param selection_method: Selection method to apply. Currently supported options are
    'random', 'maxdiag', 'maxdet', 'bait', 'fw' (for FrankWolfe), 'maxdist', 'kmeanspp', 'lcmd'
    and the experimental options 'fw-kernel' (for FrankWolfe in kernel space, slow!), 'rmds' and 'sosd' (slow!)
    :param precomp_batch_size: Batch size used for precomputations of the features.
    :param nn_batch_size: Batch size used for passing the data through the NN.
    :param config: Other options. Examples:
    allow_float64=True: enables using float64 tensors if maxdet or transformations involving posteriors are used.
    compute_eff_dim=True: Triggers the computation of the effective dimension of the pool set kernel matrix
                            for kernels with feature space dimension <= 1000.
    sel_with_train=True/False: Forces TP/P-mode for the selection method.
                                By default, the distance-based methods run in TP-mode
                                and the other ones run in P-mode.
    allow_maxdet_fs=True: Allows to compute maxdet in feature space
                            if a criterion decides that it would be sensible in terms of efficiency.
    maxdet_sigma=<float>: sigma to use in maxdet, where sigma^2 is added to the kernel diagonal.
    allow_kernel_space_posterior=False: Do not compute the posterior in kernel space
                                        unless it is strictly necessary.
                                        This can be helpful if a feature-space representation is needed,
                                        for example for FrankWolfe.
    n_last_layers=<int> (default=1): How many of the last layers to use for the 'll' base kernel.
    layer_grad_dict=<Dict[Type, Type]> (default={nn.Linear: LinearGradientComputation}):
                            Allows to specify gradient computation classes for different types of layers
                            (subclasses of nn.Module). Note that layers that inherit from LayerGradientComputation
                            will be automatically used for gradient computation.
                            By the default value of this argument, nn.Linear layers also will be used for gradients.

    There are a few other options, e.g. for the nngp and ntk kernels,
    which can be found by searching for usages of 'config' in the source code.

    :return: Returns a tuple (batch_idxs, results) where batch_idxs is a torch.Tensor of shape [batch_size]
    containing the selected indices for the pool data. The dictionary results is of the form
    {'kernel_time': {'total': <float>, 'process': <float>},
     'selection_time': {'total': <float>, 'process': <float>},
     'selection_status': <None or status message>}
    and additionally may contain 'eff_dim': <float> if compute_eff_dim=True has been passed in **config.
    Times are measured in seconds.
    """
    bs = BatchSelectorImpl(models, data, y_train)
    return bs.select(base_kernel=base_kernel, kernel_transforms=kernel_transforms, selection_method=selection_method,
                     batch_size=batch_size, precomp_batch_size=precomp_batch_size, nn_batch_size=nn_batch_size,
                     **config)


class BatchSelectorImpl:
    """
    This internal class allows to apply the methods from the paper for selecting a batch from the pool set,
    where the base kernel, kernel transformations and selection method can be easily configured.
    A new object of this class should be created every time a batch is selected,
    since the state of the class gets modified during batch selection.
    """
    def __init__(self, models: List[nn.Module], data: Dict[str, FeatureData], y_train: Optional[torch.Tensor]):
        """
        :param models: List of trained NNs. Multiple NNs can be provided if ensembling should be used,
        otherwise only one NN should be provided.
        :param data: Dict of FeatureData objects, which should be in a format
        such that the NNs can be applied to the tensors contained in the FeatureData objects.
        The dict should at least contain the keys 'train' and 'pool', for training and pool data.
        Other keys can be provided and their data will also be transformed but they are currently not used,
        so they should not be provided.
        :param y_train: Labels on the training set as a torch.Tensor of shape [n_train, 1].
        This is only used for the acs-rf-hyper transform, otherwise it is sufficient to set y_train=None.
        """
        self.data = data
        self.models = models
        self.features = {}  # will be computed in select()
        self.n_models = len(models)
        self.y_train = y_train
        self.has_select_been_called = False
        self.device = self.data['train'].get_device()

    def apply_tfm(self, model_idx: int, tfm: FeaturesTransform):
        """
        Internal method that applies a transformation to all Features objects (train/pool)
        for the model with index model_idx.
        :param model_idx: Index of the model to apply transformations to
        :param tfm: Transformation to apply to the Features.
        """
        for key in self.features:
            self.features[key][model_idx] = tfm(self.features[key][model_idx])

    def ensemble(self):
        """
        Internal method to ensemble the kernels/features for different models.
        """
        for key in self.features:
            sum_fm = SumFeatureMap([f.feature_map for f in self.features[key]])
            sum_fd = ListFeatureData([f.feature_data for f in self.features[key]])
            self.features[key] = [Features(sum_fm, sum_fd)]
        self.n_models = 1

    def to_float64(self):
        """
        Internal method to convert the data to float64.
        This only has an effect on the result if it is applied before self.features is filled.
        """
        for key in self.data:
            self.data[key] = self.data[key].cast_to(torch.float64)

    def select(self, base_kernel: str, kernel_transforms: List[Tuple[str, List]], selection_method: str,
               batch_size: int, precomp_batch_size: int = 32768, nn_batch_size=8192, **config) \
            -> Tuple[torch.Tensor, Dict[str, Any]]:
        """
        This method allows to select a batch of indices from the pool set.
        The used Batch Active Learning method can be flexibly configured.
        This method may reset the gradients of the parameters of the model(s) provided in the constructor to None.
        :param base_kernel: Base kernel to use. Currently supported base kernels are
        'll', 'grad', 'lin', 'nngp', 'ntk', and 'laplace'.
        :param kernel_transforms: List of kernel transformations.
        Each kernel transformation is given by a tuple (name, args)
        and the corresponding transformation method is then called with parameters *args.
        The following kernel transformations are currently supported:
        ('train', [sigma]) applies k_{\to\Xtrain}
        ('train', [sigma, factor]) applies (factor^2 * k)_{\to post(\Xtrain, sigma^2)},
                                    if factor is not None, or k_{\to\Xtrain} otherwise
        ('pool', [sigma]) does the same for \Xpool instead of \Xtrain
        ('pool', [sigma, factor]) does the same for \Xpool instead of \Xtrain
        ('scale', []) applies k_{\to scale(\Xtrain)}
        ('scale', [factor]) scales to factor^2 * k, if factor is not None, and k_{\to scale(\Xtrain)} otherwise
        ('rp', [n_features]) applies sketching (= random projections) with n_features random features
        ('ens', []) applies k_{\to ens}, i.e. ensembling of all kernels.
        ('acs-rf', [n_features, sigma]) applies k_{\to acs-rf(n_features)} with GP noise standard deviation sigma.
                                        As for 'train', there is also an optional third parameter
                                        for explicitly controlling the rescaling.
        ('acs-rf-hyper', [n_features]) applies k_{\to acs-rf-hyper(n_features)}.
                                        As for 'train', there is also an optional second parameter
                                        for explicitly controlling the rescaling.
        ('acs-grad', [sigma]) applies k_{\to acs-grad} with GP noise standard deviation sigma.
                                        As for 'train', there is also an optional second parameter
                                        for explicitly controlling the rescaling.
        :param selection_method: Selection method to apply. Currently supported options are
        'random', 'maxdiag', 'maxdet', 'bait', 'fw' (for FrankWolfe), 'maxdist', 'kmeanspp', 'lcmd'
        and the experimental options 'fw-kernel' (for FrankWolfe in kernel space, slow!), 'rmds' and 'sosd' (slow!)
        :param batch_size: Number of samples to select from the pool set.
        :param precomp_batch_size: Batch size used for precomputations of the features.
        :param nn_batch_size: Batch size used for passing the data through the NN.
        :param config: Other options. Examples:
        allow_float64=True: enables using float64 tensors if maxdet or transformations involving posteriors are used.
        compute_eff_dim=True: Triggers the computation of the effective dimension of the pool set kernel matrix
                                for kernels with feature space dimension <= 1000.
        sel_with_train=True/False: Forces TP/P-mode for the selection method.
                                    By default, the distance-based methods run in TP-mode
                                    and the other ones run in P-mode.
        allow_maxdet_fs=True: Allows to compute maxdet in feature space
                                if a criterion decides that it would be sensible in terms of efficiency.
        maxdet_sigma=<float>: sigma to use in maxdet, where sigma^2 is added to the kernel diagonal.
        allow_kernel_space_posterior=False: Do not compute the posterior in kernel space
                                            unless it is strictly necessary.
                                            This can be helpful if a feature-space representation is needed,
                                            for example for FrankWolfe.
        n_last_layers=<int> (default=1): How many of the last layers to use for the 'll' base kernel.
        layer_grad_dict=<Dict[Type, Type]> (default={nn.Linear: LinearGradientComputation}):
                                Allows to specify gradient computation classes for different types of layers
                                (subclasses of nn.Module). Note that layers that inherit from LayerGradientComputation
                                will be automatically used for gradient computation.
                                By the default value of this argument, nn.Linear layers also will be used for gradients.
        verbosity=<int> (default=1): Allows to control how much information will be printed.
                                     Set to a value <= 0 if no information should be printed.
        use_cuda_synchronize=True: Use CUDA synchronize for more accurate time measurements.

        There are a few other options, e.g. for the nngp and ntk kernels,
        which can be found by searching for usages of 'config' in the source code.

        :return: Returns a tuple (batch_idxs, results) where batch_idxs is a torch.Tensor of shape [batch_size]
        containing the selected indices for the pool data. The dictionary results is of the form
        {'kernel_time': {'total': <float>, 'process': <float>},
         'selection_time': {'total': <float>, 'process': <float>},
         'selection_status': <None or status message>}
         and additionally may contain 'eff_dim': <float> if compute_eff_dim=True has been passed in **config.
        """

        if self.has_select_been_called:
            raise RuntimeError('select() can only be called once per BatchSelector object')
        self.has_select_been_called = True

        allow_tf32_before = torch.backends.cuda.matmul.allow_tf32
        torch.backends.cuda.matmul.allow_tf32 = False  # do not use tf32 since it causes large numerical errors

        if config.get('allow_float64', False):
            use_float64 = (selection_method in ['maxdet', 'bait'])

            for tfm_name, tfm_args in kernel_transforms:
                if tfm_name in ['train', 'pool', 'acs-rf', 'acs-rf-hyper', 'acs-grad']:
                    use_float64 = True
        else:
            use_float64 = False

        if config.get('use_cuda_synchronize', False):
            torch.cuda.synchronize(self.device)

        kernel_timer = utils.Timer()
        kernel_timer.start()

        if base_kernel == 'ntk':  # data -> features
            feature_maps = [ReLUNTKFeatureMap(n_layers=config.get('n_ntk_layers', len([
                l for l in model.modules() if isinstance(l, LayerGradientComputation)])),
                                              sigma_w_sq=config.get('weight_gain', 0.4)**2,
                                              sigma_b_sq=config.get('sigma_b', 0.0)**2) for model in self.models]
            if use_float64:
                self.to_float64()
        elif base_kernel == 'nngp':  # data -> features
            # we use sigma_b instead of bias_gain here, since in our case, biases are initialized to zero,
            # which for the nngp corresponds to sigma_b_sq = 0 instead of sigma_b_sq = bias_gain**2
            feature_maps = [ReLUNNGPFeatureMap(n_layers=config.get('n_nngp_layers', len([
                l for l in model.modules() if isinstance(l, LayerGradientComputation)])),
                                               sigma_w_sq=config.get('weight_gain', 0.25)**2,
                                               sigma_b_sq=config.get('sigma_b', 0.0)**2) for model in self.models]
            if use_float64:
                self.to_float64()
        elif base_kernel == 'grad':  # data -> features
            feature_maps = []
            grad_dict = config.get('layer_grad_dict', {nn.Linear: LinearGradientComputation})
            for model in self.models:
                grad_layers = []
                for layer in model.modules():
                    if isinstance(layer, LayerGradientComputation):
                        grad_layers.append(layer)
                    elif type(layer) in grad_dict:
                        grad_layers.append(grad_dict[type(layer)](layer))
                feature_maps.append(create_grad_feature_map(model, grad_layers, use_float64=use_float64))
        elif base_kernel == 'll':  # data -> features
            n_last_layers = config.get('n_last_layers', 1)
            feature_maps = []
            grad_dict = config.get('layer_grad_dict', {nn.Linear: LinearGradientComputation})
            for model in self.models:
                grad_layers = []
                for layer in model.modules():
                    if isinstance(layer, LayerGradientComputation):
                        grad_layers.append(layer)
                    elif type(layer) in grad_dict:
                        grad_layers.append(grad_dict[type(layer)](layer))
                feature_maps.append(create_grad_feature_map(model, grad_layers[-n_last_layers:],
                                                            use_float64=use_float64))
        elif base_kernel == 'linear':
            feature_maps = [IdentityFeatureMap(n_features=self.data['train'].get_tensor(0).shape[-1]) for model in self.models]
            if use_float64:
                self.to_float64()
        elif base_kernel == 'laplace':
            feature_maps = [LaplaceKernelFeatureMap(scale=config.get('laplace_scale', 1.0)) for model in self.models]
            if use_float64:
                self.to_float64()
        else:
            raise ValueError(f'Unknown base kernel "{base_kernel}"')

        self.features = {key: [Features(fm, feature_data) for fm in feature_maps]
                         for key, feature_data in self.data.items()}

        if base_kernel in ['ll', 'grad']:
            for i in range(self.n_models):
                # use smaller batch size for NN evaluation
                self.apply_tfm(i, BatchTransform(batch_size=nn_batch_size))

        for tfm_name, args in kernel_transforms:
            if tfm_name == 'train':
                for i in range(self.n_models):
                    self.apply_tfm(i, PrecomputeTransform(batch_size=precomp_batch_size))
                    if len(args) >= 2:
                        self.apply_tfm(i, self.features['train'][i].scale_tfm(factor=args[1]))
                    self.apply_tfm(i, self.features['train'][i].posterior_tfm(args[0], **config))
            elif tfm_name == 'pool':
                for i in range(self.n_models):
                    self.apply_tfm(i, PrecomputeTransform(batch_size=precomp_batch_size))
                    if len(args) >= 2:
                        self.apply_tfm(i, self.features['pool'][i].scale_tfm(factor=args[1]))
                    self.apply_tfm(i, self.features['pool'][i].posterior_tfm(args[0], **config))
            elif tfm_name == 'scale':
                for i in range(self.n_models):
                    self.apply_tfm(i, PrecomputeTransform(batch_size=precomp_batch_size))
                    self.apply_tfm(i, self.features['train'][i].scale_tfm(*args))
            elif tfm_name == 'rp':
                # don't precompute before random projections
                # since we might want to jointly forward through the model and project in batches
                # otherwise we might use more memory than needed
                for i in range(self.n_models):
                    self.apply_tfm(i, self.features['train'][i].sketch_tfm(*args, **config))
            elif tfm_name == 'ens':
                self.ensemble()
            elif tfm_name == 'acs-rf':
                for i in range(self.n_models):
                    self.apply_tfm(i, PrecomputeTransform(batch_size=precomp_batch_size))
                    if len(args) >= 3:
                        self.apply_tfm(i, self.features['train'][i].scale_tfm(factor=args[2]))
                    self.apply_tfm(i, self.features['train'][i].acs_rf_tfm(args[0], args[1]))
            elif tfm_name == 'acs-rf-hyper':
                for i in range(self.n_models):
                    self.apply_tfm(i, PrecomputeTransform(batch_size=precomp_batch_size))
                    if len(args) >= 2:
                        self.apply_tfm(i, self.features['train'][i].scale_tfm(factor=args[1]))
                    if self.y_train is None:
                        raise ValueError(
                            'Set y_train to None, but y_train is needed for the acs-rf-hyper transformation')
                    self.apply_tfm(i, self.features['train'][i].acs_rf_hyper_tfm(self.y_train, n_features=args[0]))
            elif tfm_name == 'acs-grad':
                for i in range(self.n_models):
                    self.apply_tfm(i, PrecomputeTransform(batch_size=precomp_batch_size))
                    if len(args) >= 2:
                        self.apply_tfm(i, self.features['train'][i].scale_tfm(factor=args[1]))
                    self.apply_tfm(i, self.features['train'][i].acs_grad_tfm(args[0]))
            else:
                raise ValueError(f'Unknown kernel transform "{tfm_name}"')

        for i in range(self.n_models):
            self.apply_tfm(i, PrecomputeTransform(batch_size=precomp_batch_size))

        if config.get('use_cuda_synchronize', False):
            torch.cuda.synchronize(self.device)

        kernel_timer.pause()

        eff_dim = None
        if config.get('compute_eff_dim', False):
            pool_features = self.features['pool'][0]
            if pool_features.get_n_features() <= 1000:
                try:
                    feature_matrix = pool_features.get_feature_matrix()
                    feature_matrix = feature_matrix - feature_matrix.mean(dim=0, keepdim=True)
                    cov_matrix = feature_matrix.t().matmul(feature_matrix)
                    cov_matrix = 0.5 * (cov_matrix + cov_matrix.t())
                    largest_eigval = torch.linalg.matrix_norm(cov_matrix, 2)
                    eff_dim = (torch.trace(cov_matrix) / (torch.abs(largest_eigval) + 1e-30)).item()
                except Exception as e:
                    pass

        if config.get('use_cuda_synchronize', False):
            torch.cuda.synchronize(self.device)

        selection_timer = utils.Timer()
        selection_timer.start()

        # only pick first model (if multiple models were there, they should have been ensembled by now)
        self.features = {key: val[0] for key, val in self.features.items()}

        if selection_method == 'random':
            alg = RandomSelectionMethod(self.features['pool'], **config)
        elif selection_method == 'maxdiag':
            alg = MaxDiagSelectionMethod(self.features['pool'], **config)
        elif selection_method == 'maxdet':
            sel_with_train = config.get('sel_with_train', None)
            n_select = batch_size
            n_features = self.features['pool'].get_n_features()
            maxdet_sigma = config.get('maxdet_sigma', 0.0)
            if sel_with_train:
                n_select += self.features['train'].get_n_samples()
            if config.get('allow_maxdet_fs', False) and n_features > 0 and n_features * 4 < n_select \
                    and maxdet_sigma > 0.0:
                alg = MaxDetFeatureSpaceSelectionMethod(self.features['pool'], self.features['train'],
                                                        noise_sigma=maxdet_sigma, **config)
            else:
                alg = MaxDetSelectionMethod(self.features['pool'], self.features['train'],
                                            noise_sigma=config.get('maxdet_sigma', 0.0), **config)
        elif selection_method == 'bait':
            alg = BaitFeatureSpaceSelectionMethod(self.features['pool'], self.features['train'],
                                            noise_sigma=config.get('bait_sigma', 0.0), **config)
        elif selection_method == 'maxdist':
            alg = MaxDistSelectionMethod(self.features['pool'], self.features['train'], **config)
        elif selection_method == 'lcmd':
            alg = LargestClusterMaxDistSelectionMethod(self.features['pool'], self.features['train'], **config)
        elif selection_method == 'rmds':
            alg = RandomizedMinDistSumSelectionMethod(self.features['pool'], self.features['train'],
                                                      max_n_candidates=config.get('max_rmds_candidates', 5), **config)
        elif selection_method == 'fw':
            alg = FrankWolfeSelectionMethod(self.features['pool'], self.features['train'], **config)
        elif selection_method == 'fw-kernel':
            # scales quadratically with pool set size
            alg = FrankWolfeKernelSpaceSelectionMethod(self.features['pool'], self.features['train'], **config)
        elif selection_method == 'kmeanspp':
            alg = KmeansppSelectionMethod(self.features['pool'], self.features['train'], **config)
        elif selection_method == 'sosd':
            # scales quadratically with pool set size
            alg = SumOfSquaredDistsSelectionMethod(self.features['pool'], self.features['train'], **config)
        else:
            raise ValueError(f'Unknown selection method "{selection_method}"')

        batch_idxs = alg.select(batch_size)

        if config.get('use_cuda_synchronize', False):
            torch.cuda.synchronize(self.device)

        selection_timer.pause()

        results_dict = {'kernel_time': kernel_timer.get_result_dict(),
                        'selection_time': selection_timer.get_result_dict(),
                        'selection_status': alg.get_status()}

        if eff_dim is not None:
            results_dict['eff_dim'] = eff_dim

        torch.backends.cuda.matmul.allow_tf32 = allow_tf32_before

        return batch_idxs, results_dict



import torch.nn as nn

from .feature_maps import *


class LayerGradientComputation:
    """
    Abstract base class that can be used as a second base class
    for layers that support the computation of gradient features
    """
    def __init__(self, layer: nn.Module):
        """
        :param layer: Layer that this LayerGradientComputation computes gradients of.
        """
        self.layers = [layer]  # dirty hack to avoid infinite recursion in PyTorch if layer is self.
        super().__init__()   # in case this is used with multiple inheritance

    def get_layer(self) -> nn.Module:
        """
        :return: Returns the layer that this LayerGradientComputation computes gradients of.
        """
        return self.layers[0]

    def get_feature_map(self) -> FeatureMap:
        """
        :return: Returns a FeatureMap object that can compute feature map / kernel values
        on the data provided by pop_feature_data()
        """
        raise NotImplementedError()

    def before_forward(self) -> None:
        """
        Callback that is called before the data is passed through the model in a forward pass
        and gradients are computed in a backward pass.
        This method can be used to set up hooks that grab input data and gradients in both forward and backward pass.
        """
        raise NotImplementedError()

    def pop_feature_data(self) -> FeatureData:
        """
        :return: This method should return the feature data
        corresponding to the inputs that were last passed through the model.
        This feature data should be usable by the feature map returned by get_feature_map()
        """
        raise NotImplementedError()


class ModelGradTransform(DataTransform):
    """
    A DataTransform object that passes data through a NN model
    in order to obtain feature data corresponding to gradients
    """
    def __init__(self, model: nn.Module, grad_layers: List[LayerGradientComputation]):
        """
        :param model: The model to be computed gradients of
        :param grad_layers: All layers of the model whose parameters we want to compute gradients of
        """
        self.model = model
        self.grad_layers = grad_layers
        self.requires_grad_list = [any([any([p is gl_p for gl_p in gl.get_layer().parameters()]) for gl in grad_layers])
                                   for p in model.parameters()]
        self.grad_params = [p for grad_layer in grad_layers for p in grad_layer.get_layer().parameters()]

    def forward(self, feature_data: FeatureData, idxs: Indexes) -> FeatureData:
        """
        :param feature_data: Feature data to be passed through the model
        :param idxs: indexes of the feature data that should be passed through the model
        :return: feature data provided by the layers
        """
        for grad_layer in self.grad_layers:
            grad_layer.before_forward()

        # only set requires_grad=True for those parameters that need one
        requires_grad_before_list = [p.requires_grad for p in self.model.parameters()]
        for p, requires_grad in zip(self.model.parameters(), self.requires_grad_list):
            p.requires_grad = requires_grad

        old_training = self.model.training
        self.model.eval()
        X = feature_data.get_tensor(idxs)
        y = self.model(X)  # implicitly calls hooks that were set by l.before_forward()
        y.backward(torch.ones_like(y))
        with torch.no_grad():
            for p in self.model.parameters():
                p.grad = None

        self.model.train(old_training)

        data = ListFeatureData([layer_comp.pop_feature_data() for layer_comp in self.grad_layers])

        for p, value in zip(self.model.parameters(), requires_grad_before_list):
            p.requires_grad = value

        return data


def create_grad_feature_map(model: nn.Module, grad_layers: List[LayerGradientComputation],
                            use_float64: bool = False) -> FeatureMap:
    """
    Creates a feature map corresponding to phi_{grad} or phi_{ll}, depending on which layers are provided.
    :param model: Model to compute gradients of
    :param grad_layers: All layers of the model whose parameters we want to compute gradients of
    :param use_float64: Set to true if the gradient features should be converted to float64 after computing them
    :return: Returns a feature map corresponding to phi_{grad} for the given layers.
    """
    tfms = [ModelGradTransform(model, grad_layers)]
    if use_float64:
        tfms.append(ToDoubleTransform())
    return SequentialFeatureMap(SumFeatureMap([l.get_feature_map() for l in grad_layers]),
                                tfms)


# ----- Specific LayerGradientComputation implementation(s) for linear layers


class GeneralLinearGradientComputation(LayerGradientComputation):
    """
    Implements LayerGradientFeatures for general linear layers.
    It can also be used with the Neural Tangent Parameterization since it includes a weight factor and bias factor.
    (These are called sigma_w and sigma_b in the paper.)
    """
    def __init__(self, layer: nn.Module, in_features: int, out_features: int,
                 weight_factor: float = 1.0, bias_factor: float = 1.0):
        """
        :param layer: nn.Module object implementing a linear (fully-connected) layer,
        whose gradients should be computed.
        :param in_features: Input dimension of the layer.
        :param out_features: Output dimension of the layer.
        :param weight_factor: Factor sigma_w by which the weight matrix is multiplied in the forward pass.
        :param bias_factor: Factor sigma_w by which the bias is multiplied in the forward pass.
        """
        super().__init__(layer)
        self.in_features = in_features
        self.out_features = out_features
        self.weight_factor = weight_factor
        self.bias_factor = bias_factor
        self._input_data = None
        self._grad_output_data = None
        self._input_hook = None
        self._grad_output_hook = None

    def get_feature_map(self) -> FeatureMap:
        # gradients wrt to this layer are an outer product of the input and the output gradient,
        # so we can use a ProductFeatureMap
        # the +1 is for the bias
        return ProductFeatureMap([IdentityFeatureMap(n_features=self.in_features+1),
                                  IdentityFeatureMap(n_features=self.out_features)])

    def set_input_(self, value: torch.Tensor):
        # this is used to have a method to call in the hooks
        self._input_data = value

    def set_grad_output_(self, value: torch.Tensor):
        # this is used to have a method to call in the hooks
        self._grad_output_data = value

    def before_forward(self):
        # sets up hooks that store the input and grad_output
        self._input_hook = self.get_layer().register_forward_hook(
                lambda layer, inp, output, s=self: s.set_input_(inp[0].detach().clone()))
        self._grad_output_hook = self.get_layer().register_full_backward_hook(
                lambda layer, grad_input, grad_output, s=self: s.set_grad_output_(grad_output[0].detach().clone()))

    def pop_feature_data(self) -> FeatureData:
        # remove the hooks
        self._input_hook.remove()
        self._grad_output_hook.remove()
        # compute the adjusted input \tilde{x} from the paper
        inp = torch.cat([self.weight_factor * self._input_data,
                         self.bias_factor * torch.ones(self._input_data.shape[0], 1, device=self._input_data.device)],
                        dim=1)
        # feature data for the two IdentityFeatureMaps in the ProductFeatureMap, given by inputs and grad_outputs
        fd = ListFeatureData([TensorFeatureData(inp), TensorFeatureData(self._grad_output_data)])

        # allow to release memory earlier
        self._input_data = None
        self._grad_output_data = None

        return fd


class LinearGradientComputation(GeneralLinearGradientComputation):
    """
    This class implements a gradient computation for nn.Linear layers.
    """
    def __init__(self, layer: nn.Linear):
        super().__init__(layer=layer, in_features=layer.in_features, out_features=layer.out_features)


class LinearLayer(GeneralLinearGradientComputation, nn.Module):
    """
    Linear layer that implements LayerGradientFeatures, i.e., it can be used for computing gradient-based kernels.
    This linear layer does not initialize weight and bias itself,
    instead it assumes that they are passed as arguments to the constructor.
    It can also be used with the Neural Tangent Parameterization since it includes a weight factor and bias factor.
    (These are called sigma_w and sigma_b in the paper.)
    """
    def __init__(self, weight: torch.Tensor, bias: torch.Tensor, weight_factor: float, bias_factor: float):
        """
        :param weight: Weight matrix parameter of shape [in_features, out_features].
        Compared to torch.nn.Linear, this is transposed.
        :param bias: Bias parameter of shape [out_features]
        :param weight_factor: Factor sigma_w by which the weight matrix is multiplied in the forward pass.
        :param bias_factor: Factor sigma_w by which the bias is multiplied in the forward pass.
        """
        super().__init__(self, in_features=weight.shape[0], out_features=weight.shape[1],
                         weight_factor=weight_factor, bias_factor=bias_factor)
        self.weight = nn.Parameter(weight)
        self.bias = nn.Parameter(bias)
        self.weight_factor = weight_factor
        self.bias_factor = bias_factor

    def forward(self, x: torch.Tensor):
        return self.weight_factor * x.matmul(self.weight) + self.bias_factor * self.bias







from .layers import *


def create_tabular_model(n_models, n_features, hidden_sizes=[512]*2, act='relu', n_outputs: int = 1, **config):
    layer_sizes = [n_features] + hidden_sizes + [n_outputs]
    layers = []
    for in_features, out_features in zip(layer_sizes[:-2], layer_sizes[1:-1]):
        layers.append(ParallelLinearLayer(n_models, in_features, out_features, **config))
        layers.append(get_parallel_act_layer(act))
    layers.append(ParallelLinearLayer(n_models, layer_sizes[-2], layer_sizes[-1],
                                      weight_init_mode='zero' if config.get('use_llz', False) else 'normal', **config))
    return ParallelSequential(*layers)




import matplotlib
#matplotlib.use('Agg')
matplotlib.use('pdf')
matplotlib.rcParams.update({
    "pgf.texsystem": "pdflatex",
    'font.family': 'serif',
    'font.size': 10.95,
    'text.usetex': True,
    'pgf.rcfonts': False,
    # 'legend.framealpha': 0.5,
    'text.latex.preamble': r'\usepackage{times} \usepackage{amsmath} \usepackage{amsfonts} \usepackage{amssymb}'
})
# from tueplots import bundles, fonts, fontsizes, figsizes
# matplotlib.rcParams.update(bundles.jmlr2001())
# matplotlib.rcParams.update(fonts.jmlr2001_tex())
# matplotlib.rcParams.update(fontsizes.jmlr2001())

import matplotlib.pyplot as plt
from .analysis import *
from pathlib import Path
import seaborn as sns

fontsize = 10
axis_font = {'size': str(fontsize)}
sns.axes_style("whitegrid")


def escape(name: str):
    return name.replace('_', r'\_')


def rgb_to_hex(rgb):
    return '#%02x%02x%02x' % rgb


colors = [u'#a06010', u'#d62728', u'#e377c2', u'#2ca02c', u'#ff7f0e', u'#9467bd', u'#17becf', u'#7f7f7f',
          u'#bcbd22', u'#1f77b4']
markers = ['^', 'v', '<', '>', 'P', 'D', 's', 'v']


def plot_batch_sizes_ax(ax: plt.Axes, results: ExperimentResults, metric_name: str, set_ticks_and_labels: bool = True,
                        **plot_options):
    last_errors = results.get_last_errors(metric_name)
    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in results.task_names})

    # https://stackoverflow.com/questions/42086276/get-default-line-colour-cycle
    # default matplotlib colors:
    # colors = [u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f',
    #           u'#bcbd22', u'#17becf']

    color_idx = 0

    plot_options = utils.update_dict(dict(alpha=1.0, markersize=3.5), plot_options)

    for alg_name in results.alg_names:
        log_means = []
        alg_stds = []

        ds_batch_sizes = []

        for ds_name in ds_names:
            keys = []
            batch_sizes = []

            for key in last_errors.results_dict[alg_name]:
                if key.startswith(ds_name):
                    keys.append(key)
                    batch_sizes.append(int(key.split('_')[-1].split('x')[0]))

            batch_sizes = np.array(list(batch_sizes))
            idxs = np.argsort(batch_sizes)
            batch_sizes = batch_sizes[idxs]
            results_list = [np.mean(last_errors.results_dict[alg_name][key]) for key in keys]
            task_stds = [np.std(last_errors.results_dict[alg_name][key], axis=0) /
                        np.sqrt(len(last_errors.results_dict[alg_name][key])-1) for key in keys]
            alg_stds.append(np.asarray(task_stds)[idxs])
            log_means.append(np.asarray(results_list)[idxs])
            ds_batch_sizes.append(batch_sizes)

        if len(ds_batch_sizes) == 0:
            raise ValueError(f'No data set results available for alg {alg_name}')

        if alg_name == 'NN_random':
            # all batch sizes should provide equivalent results
            minus = np.mean(log_means) - np.linalg.norm(alg_stds) / len(alg_stds)
            plus = np.mean(log_means) + np.linalg.norm(alg_stds) / len(alg_stds)
            ax.axhspan(minus, plus, alpha=0.2, facecolor='k', edgecolor=None)
            ax.axhline(y=np.mean(log_means), ls='--', color='k',
                        label=get_latex_selection_method(alg_name.split('_')[1]))
        else:
            minus = np.mean(log_means, axis=0) - np.linalg.norm(alg_stds, axis=0) / len(alg_stds)
            plus = np.mean(log_means, axis=0) + np.linalg.norm(alg_stds, axis=0) / len(alg_stds)
            ax.fill_between(np.log(ds_batch_sizes[0]), minus, plus, facecolor=colors[color_idx], alpha=0.2)
            ax.plot(np.log(ds_batch_sizes[0]), np.mean(log_means, axis=0), '--', marker=markers[color_idx],
                    label=get_latex_selection_method(alg_name.split('_')[1]), color=colors[color_idx], **plot_options)
            color_idx += 1

    if set_ticks_and_labels:
        xlocs = [np.log(64), np.log(128), np.log(256), np.log(512), np.log(1024), np.log(2048), np.log(4096)]
        xlabels = ('64', '128', '256', '512', '1024', '2048', '4096')

        ax.set_xticks(xlocs)
        ax.set_xticklabels(xlabels)

        ax.set_xlabel(r'Acquisition batch size $N_{\mathrm{batch}}$', **axis_font)
        ax.set_ylabel(r'mean log ' + f'{get_latex_metric_name(metric_name)}', **axis_font)


def plot_batch_sizes(results: ExperimentResults, filename: Union[str, Path], metric_name: str,
                     figsize: Optional[Tuple[float, float]] = None):
    # plot mean final log metric against the al batch size
    fig, ax = plt.subplots(figsize=figsize or (4, 4))

    plot_batch_sizes_ax(ax, results, metric_name)

    ax.legend()
    plt.tight_layout()
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name)
    plt.close(fig)


def plot_multiple_batch_sizes(results: ExperimentResults, filename: Union[str, Path],
                                  metric_names: List[str]):
    # plot averaged learning curve for all tasks
    fig, axs = plt.subplots(1, len(metric_names), figsize=(7, 4))

    for i, metric_name in enumerate(metric_names):
        plot_batch_sizes_ax(axs[i], results, metric_name)

    fig.legend(*axs[0].get_legend_handles_labels(), loc='upper center', bbox_to_anchor=(0.5, 0.15), ncol=4)
    plt.tight_layout(rect=[0, 0.15, 1.0, 1.0])
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name)
    plt.close(fig)


def plot_batch_sizes_individual(results: ExperimentResults, metric_name: str):
    # plot mean final log metric against the al batch size
    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in results.task_names})
    ds_names.sort()

    for ds_name in ds_names:
        fig, ax = plt.subplots(figsize=(4, 4))
        plot_batch_sizes_ax(ax, results.filter_task_prefix(ds_name), metric_name)

        ax.legend()
        plt.tight_layout()
        plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / 'ds_batch_sizes' / \
                    f'{ds_name}_{metric_name}.pdf'
        utils.ensureDir(plot_name)
        plt.savefig(plot_name)
        plt.close(fig)


def plot_batch_sizes_individual_subplots(results: ExperimentResults, filename: Union[str, Path], metric_name: str):
    # plot mean final log metric against the al batch size
    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in results.task_names})
    ds_names.sort()

    fig, axs = plt.subplots(5, 3, figsize=(9, 12))
    if len(ds_names) != 15:
        print(f'plot_batch_sizes_individual_subplots needs 15 data sets, but got {len(ds_names)} data sets')
        return
    for ds_idx, ds_name in enumerate(ds_names):
        i = ds_idx // 3
        j = ds_idx % 3
        ax = axs[i, j]
        plot_batch_sizes_ax(ax, results.filter_task_prefix(ds_name), metric_name, set_ticks_and_labels=False,
                            markersize=3.5)

        xlocs = [np.log(64), np.log(128), np.log(256), np.log(512), np.log(1024), np.log(2048), np.log(4096)]
        xlabels = ('64', '128', '256', '512', '1024', '2048', '4096')

        ax.set_xticks(xlocs)
        ax.set_xticklabels(xlabels if i == 4 else [''] * len(xlocs))

        if i == 4:
            ax.set_xlabel(r'Acquisition batch size $N_{\mathrm{batch}}$', **axis_font)
        if j == 0:
            ax.set_ylabel(r'mean log ' + f'{get_latex_metric_name(metric_name)}', **axis_font)
        ax.set_title(get_latex_ds_name(ds_name))

    fig.legend(*axs[4, 1].get_legend_handles_labels(), loc='upper center', bbox_to_anchor=(0.5, 0.05), ncol=4)
    plt.tight_layout(rect=[0, 0.05, 1.0, 1.0])
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name)
    plt.close(fig)


def plot_batch_sizes_metrics_subplots(results: ExperimentResults, filename: Union[str, Path]):
    # plot mean final log metric against the al batch size

    fig, axs = plt.subplots(3, 2, figsize=(8, 9))
    metric_names = ['mae', 'rmse', 'q95', 'q99', 'maxe']
    for metric_idx, metric_name in enumerate(metric_names):
        i = metric_idx // 2
        j = metric_idx % 2
        ax = axs[i, j]
        plot_batch_sizes_ax(ax, results, metric_name, set_ticks_and_labels=True)

    axs[-1, -1].axis('off')
    fig.legend(*axs[0, 0].get_legend_handles_labels(), loc='center', bbox_to_anchor=(0.78, 0.18), ncol=1)
    plt.tight_layout()
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name)
    plt.close(fig)


def plot_learning_curves_ax(ax: plt.Axes, results: ExperimentResults, metric_name: str, with_random_final: bool = True,
                            set_ticks_and_labels: bool = True, labels: Optional[List[str]] = None, **plot_options):
    learning_curves = results.get_learning_curves(metric_name)

    # https://stackoverflow.com/questions/42086276/get-default-line-colour-cycle
    # default matplotlib colors:
    # colors = [u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f',
    #           u'#bcbd22', u'#17becf']
    # colors = [u'#a06010', u'#d62728', u'#2ca02c', u'#ff7f0e', u'#9467bd', u'#17becf', u'#e377c2', u'#7f7f7f',
    #           u'#bcbd22', u'#1f77b4']

    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in results.task_names})

    color_idx = 0

    for i, alg_name in enumerate(results.alg_names):
        log_means = [learning_curves.results_dict[alg_name][ds_name + '_256x16'] for ds_name in ds_names]
        results_list = np.mean(log_means, axis=0)
        n_train = np.asarray([256*(i+1) for i in range(len(results_list))])

        plot_options = utils.update_dict(dict(alpha=1.0, markersize=3.5), plot_options)

        label = get_latex_selection_method(alg_name.split('_')[1]) if labels is None else labels[i]

        if alg_name == 'NN_random':
            ax.plot(np.log(n_train), results_list, '--o', color='k', label=label, **plot_options)
            if with_random_final:
                ax.plot([np.log(n_train[0]), np.log(n_train[-1])], [results_list[-1], results_list[-1]],
                        '--', color='k', **plot_options)
        else:
            ax.plot(np.log(n_train), results_list, '--', marker=markers[color_idx], color=colors[color_idx],
                    label=label, **plot_options)
            color_idx += 1

    if set_ticks_and_labels:
        xlocs = (np.log(256), np.log(512), np.log(1024), np.log(2048), np.log(4096))
        xlabels = ('256', '512', '1024', '2048', '4096')

        ax.set_xticks(xlocs)
        ax.set_xticklabels(xlabels)

        ax.set_xlabel(r'Training set size $N_{\mathrm{train}}$', **axis_font)
        ax.set_ylabel(r'mean log ' + f'{get_latex_metric_name(metric_name)}', **axis_font)


def plot_learning_curves(results: ExperimentResults, filename: Union[str, Path], metric_name: str,
                         labels: Optional[List[str]] = None, figsize: Optional[Tuple[float, float]] = None):
    # plot averaged learning curve for all tasks
    fig, axs = plt.subplots(figsize=figsize or (3.5, 3.5))

    plot_learning_curves_ax(axs, results=results, metric_name=metric_name, labels=labels)

    axs.legend()
    plt.tight_layout()
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name)
    plt.close(fig)


def plot_multiple_learning_curves(results: ExperimentResults, filename: Union[str, Path],
                                  metric_names: List[str]):
    # plot averaged learning curve for all tasks
    fig, axs = plt.subplots(1, len(metric_names), figsize=(7, 4))

    for i, metric_name in enumerate(metric_names):
        plot_learning_curves_ax(axs[i], results, metric_name=metric_name)

    fig.legend(*axs[0].get_legend_handles_labels(), loc='upper center', bbox_to_anchor=(0.5, 0.15), ncol=4)
    plt.tight_layout(rect=[0, 0.15, 1.0, 1.0])
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name)
    plt.close(fig)


def plot_learning_curves_individual(results: ExperimentResults, metric_name: str):
    # plot individual learning curve for all tasks
    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in results.task_names})
    ds_names.sort()

    for ds_name in ds_names:
        fig, axs = plt.subplots(figsize=(3.5, 3.5))
        plot_learning_curves_ax(axs, results.filter_task_names([ds_name + '_256x16']), metric_name=metric_name)
        axs.legend(fontsize=fontsize)
        plt.tight_layout()
        plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / 'ds_learning_curves' / \
                    f'{ds_name}_{metric_name}.pdf'
        utils.ensureDir(plot_name)
        plt.savefig(plot_name)
        plt.close(fig)


def plot_learning_curves_individual_subplots(results: ExperimentResults, filename: str, metric_name: str):
    # plot individual learning curve for all tasks
    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in results.task_names})
    ds_names.sort()

    fig, axs = plt.subplots(5, 3, figsize=(9, 12))

    if len(ds_names) != 15:
        print(f'plot_batch_sizes_individual_subplots needs 15 data sets, but got {len(ds_names)} data sets')
        return
    for ds_idx, ds_name in enumerate(ds_names):
        i = ds_idx // 3
        j = ds_idx % 3
        ax = axs[i, j]
        plot_learning_curves_ax(ax, results.filter_task_names([ds_name + '_256x16']), metric_name=metric_name,
                                set_ticks_and_labels=False, alpha=0.8, markersize=3, linewidth=1.0, markeredgewidth=0.0)

        xlocs = [np.log(256), np.log(512), np.log(1024), np.log(2048), np.log(4096)]
        xlabels = ('256', '512', '1024', '2048', '4096')

        ax.set_xticks(xlocs)
        ax.set_xticklabels(xlabels if i == 4 else [''] * len(xlocs))

        if i == 4:
            ax.set_xlabel(r'Training set size $N_{\mathrm{train}}$', **axis_font)
        if j == 0:
            ax.set_ylabel(f'mean log {get_latex_metric_name(metric_name)}', **axis_font)
        ax.set_title(get_latex_ds_name(ds_name))

    fig.legend(*axs[4, 1].get_legend_handles_labels(), loc='upper center', bbox_to_anchor=(0.5, 0.05), ncol=4)
    plt.tight_layout(rect=[0, 0.05, 1.0, 1.0])

    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name)
    plt.close(fig)


def plot_learning_curves_metrics_subplots(results: ExperimentResults, filename: Union[str, Path]):
    # plot mean final log metric against the al batch size
    fig, axs = plt.subplots(3, 2, figsize=(7.2, 9))
    metric_names = ['mae', 'rmse', 'q95', 'q99', 'maxe']
    for metric_idx, metric_name in enumerate(metric_names):
        i = metric_idx // 2
        j = metric_idx % 2
        ax = axs[i, j]
        plot_learning_curves_ax(ax, results, metric_name)

    axs[-1, -1].axis('off')
    fig.legend(*axs[0, 0].get_legend_handles_labels(), loc='center', bbox_to_anchor=(0.78, 0.18), ncol=1)
    plt.tight_layout()
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name)
    plt.close(fig)


def plot_correlation_between_methods(results: ExperimentResults, filename: Union[str, Path], metric_name: str):
    sns.color_palette("Paired")

    # plot mean final log metric against the al batch size
    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in results.task_names})
    ds_names.sort()
    task_names = [ds_name + '_256x16' for ds_name in ds_names]
    alg_names = [alg_name for alg_name in results.alg_names if alg_name != 'NN_random']

    fig, axs = plt.subplots(len(alg_names)-1, len(alg_names)-1, figsize=(7.8, 7.8), sharex='col', sharey='row')

    last_errors = results.get_avg_errors(metric_name)

    all_results = {alg_name: {task_name: np.mean(last_errors.results_dict[alg_name][task_name])
                              for task_name in task_names} for alg_name in alg_names + ['NN_random']}

    max_result = 0.0
    min_result = 0.0
    alpha = 0.7
    markersize = 6

    # generated using https://github.com/taketwo/glasbey
    # c_list = [(0,0,0), (215,0,0), (140,60,255), (2,136,0), (0,172,199), (152,255,0), (255,127,209), (108,0,79), (255,165,48),
    #      (0,0,157), (134,112,104), (0,73,66), (79,42,0), (0,253,207), (188,183,255)]
    c_list = [(174,20,20), (0,85,239), (0,143,0), (239,91,255), (225,149,10), (0,184,196), (120,78,120), (255,110,128),
              (114,90,36), (148,155,255), (19,109,100), (145,180,125), (148,55,252), (202,16,130), (97,122,157)]

    c = [rgb_to_hex(c) for c in c_list]

    for i in range(len(alg_names)):
        for j in range(i+1, len(alg_names)):
            alg_name_i = alg_names[i]
            alg_name_j = alg_names[j]
            k = 0
            for task_name in task_names:
                result_x = all_results[alg_name_i][task_name] - all_results['NN_random'][task_name]
                result_y = all_results[alg_name_j][task_name] - all_results['NN_random'][task_name]
                max_result = np.max([result_x, result_y, max_result])
                min_result = np.min([result_x, result_y, min_result])
                if i == 0 and j == 1:
                    axs[j - 1, i].plot(result_x, result_y, 'o', alpha=alpha,
                                       label=get_latex_task(task_name.split('_256')[0]), markersize=markersize,
                                       color=c[k], markeredgewidth=0.0)
                    k += 1
                    handles, labels = axs[0, 0].get_legend_handles_labels()
                    axs[0, -1].legend(handles, labels, loc='upper right', borderaxespad=0, fontsize=fontsize)
                else:
                    axs[j - 1, i].plot(result_x, result_y, 'o', markersize=markersize, alpha=alpha, color=c[k],
                                       markeredgewidth=0.0)
                    k += 1

                if i == 0:
                    axs[j - 1, i].set_ylabel(get_latex_selection_method(alg_name_j.split('_')[1]), **axis_font)
                if j == len(alg_names)-1:
                    axs[j - 1, i].set_xlabel(get_latex_selection_method(alg_name_i.split('_')[1]), **axis_font)

    for i in range(len(alg_names)):
        for j in range(i+1, len(alg_names)):
            axs[j-1, i].plot([min_result, max_result], [min_result, max_result], 'k-')
            axs[j-1, i].plot([min_result, max_result], [0.0, 0.0], 'k--')
            axs[j - 1, i].plot([0.0, 0.0], [min_result, max_result], 'k--')

    for i in range(len(alg_names)-1):
        for j in range(i+1, len(alg_names)-1):
            axs[i, j].axis('off')

    plt.subplots_adjust(wspace=0.05, hspace=0.05)
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name, bbox_inches="tight")
    plt.close(fig)


def plot_skewness_ax(ax: plt.Axes, results: ExperimentResults, metric_name: str, alg_name: str,
                     use_relative_improvement: bool = False):
    sns.color_palette("Paired")

    # plot mean final log metric against the al batch size
    all_alg_names = list(results.results_dict.keys())
    all_task_names = {task_name for alg_name in all_alg_names for task_name in results.results_dict[alg_name].keys()}
    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in all_task_names})
    task_names = [ds_name + '_256x16' for ds_name in ds_names]
    task_names.sort()

    avg_errors = results.get_avg_errors(metric_name)
    random_name = 'NN_random'

    all_results = {name: {task_name: np.mean(avg_errors.results_dict[name][task_name])
                              for task_name in task_names} for name in [alg_name, random_name]}

    alpha = 0.7
    markersize = 6

    # generated using https://github.com/taketwo/glasbey
    # c_list = [(0,0,0), (215,0,0), (140,60,255), (2,136,0), (0,172,199), (152,255,0), (255,127,209), (108,0,79), (255,165,48),
    #      (0,0,157), (134,112,104), (0,73,66), (79,42,0), (0,253,207), (188,183,255)]
    c_list = [(174,20,20), (0,85,239), (0,143,0), (239,91,255), (225,149,10), (0,184,196), (120,78,120), (255,110,128),
              (114,90,36), (148,155,255), (19,109,100), (145,180,125), (148,55,252), (202,16,130), (97,122,157)]

    c = [rgb_to_hex(c) for c in c_list]

    x_values = []
    y_values = []

    for task_idx, task_name in enumerate(task_names):
        random_results = results.results_dict[random_name][task_name]
        n_splits = len(random_results)
        first_step_results = {metric: np.mean([np.log(random_results[i]['errors'][metric][0]) for i in range(n_splits)])
                              for metric in ['rmse', 'mae', 'q95', 'q99', 'maxe']}
        x_value = first_step_results['rmse'] - first_step_results['mae']
        y_value = all_results[alg_name][task_name] - all_results[random_name][task_name]
        if use_relative_improvement:
            # divide by (avg log rmse - initial log rmse)
            diff = all_results[random_name][task_name] - first_step_results[metric_name]
            y_value /= diff
        x_values.append(x_value)
        y_values.append(y_value)
        ax.plot(x_value, y_value, 'o', alpha=alpha, label=get_latex_task(task_name.split('_256')[0]),
                markersize=markersize, color=c[task_idx], markeredgewidth=0.0)
        sel_method_name = get_latex_selection_method(alg_name.split('_')[1])
        latex_metric_name = get_latex_metric_name(metric_name)
        if use_relative_improvement:
            ax.set_ylabel(f'Relative improvement of {sel_method_name}', **axis_font)
        else:
            ax.set_ylabel(f'Relative mean log {latex_metric_name}', **axis_font)
        ax.set_xlabel(f'mean log RMSE - mean log MAE', **axis_font)

    x_min, x_max = np.min(x_values), np.max(x_values)
    x_min = 0.0  # extend to 0.0
    ax.plot([x_min, x_max], [0.0, 0.0], 'k--', label=r'\textsc{Random}')

    # see https://stackoverflow.com/questions/893657/how-do-i-calculate-r-squared-using-python-and-numpy
    import scipy.stats
    slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(x_values, y_values)
    ax.plot([x_min, x_max], [slope * x_min + intercept, slope * x_max + intercept], '-', color='#888888',
            label='Linear Regression fit')
    print(f'plot_skewness_ax: R^2 = {r_value**2:g}')


def plot_error_variation(results: ExperimentResults, filename: Union[str, Path], metric_name: str, alg_name: str,
                         use_relative_improvement: bool = False):
    fig, ax = plt.subplots(1, 1, figsize=(4.0, 4.0))

    plot_skewness_ax(ax, results, metric_name=metric_name, alg_name=alg_name,
                     use_relative_improvement=use_relative_improvement)
    plt.legend(loc='upper left', bbox_to_anchor=(1.05, 1.02))
    # plt.tight_layout()
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name, bbox_inches="tight")
    plt.close(fig)


# ----- not needed plots -----


def plot_eff_dim(results: ExperimentResults, filename: Union[str, Path], pattern: str, metric_name: str):
    # plots, for every data set, error of grad_rp-512 to ll vs difference in effective dimensions (average),
    # for different selection methods (esp. maxdist, maxdet, lcmd, fw),
    all_alg_names = [pattern.replace('*', 'll'), pattern.replace('*', 'grad_rp-512')]
    all_task_names = {task_name for alg_name in all_alg_names for task_name in results.results_dict[alg_name].keys()}

    eff_dims = results.get_avg_al_stats('eff_dim')
    last_errors = results.get_last_errors(metric_name)
    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in all_task_names})

    fig, axs = plt.subplots(figsize=(5, 5))

    for name_task in ds_names:
        log_means_ll = []
        log_means_grad_rp = []
        for key in last_errors.results_dict[all_alg_names[0]]:
            if key.startswith(name_task) and int(key.split('_')[-1].split('x')[0]) == 256:
                log_means_ll = [np.mean(last_errors.results_dict[all_alg_names[0]][key]), np.mean(eff_dims.results_dict[all_alg_names[0]][key])]
                log_means_grad_rp = [np.mean(last_errors.results_dict[all_alg_names[1]][key]), np.mean(eff_dims.results_dict[all_alg_names[1]][key])]

        log_means_ll = np.asarray(log_means_ll)
        log_means_grad_rp = np.asarray(log_means_grad_rp)

        axs.plot(log_means_grad_rp[1] - log_means_ll[1], log_means_ll[0] - log_means_grad_rp[0], 'o', label=escape(name_task))

    axs.legend()
    plt.tight_layout()
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name)
    plt.close(fig)


def plot_learning_curves_v2(results: ExperimentResults, filename: Union[str, Path], metric_name: str):
    # plot averaged learning curve for all tasks
    all_alg_names = list(results.results_dict.keys())
    all_task_names = {task_name for alg_name in all_alg_names for task_name in results.results_dict[alg_name].keys()}
    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in all_task_names})
    algs_with_all_tasks = [alg_name for alg_name in all_alg_names
                           if set(results.results_dict[alg_name].keys()) == all_task_names]
    alg_names = algs_with_all_tasks + ['NN_lcmd-tp_ll', 'NN_kmeanspp-tp_ll', 'NN_maxdist-tp_grad_rp-512', 'NN_random']

    fig, axs = plt.subplots(figsize=(5, 5))
    plt.xscale('log')

    learning_curves = results.get_learning_curves(metric_name)

    n_train = np.asarray([256 * (i + 1) for i in range(17)])
    alg_results = {alg_name: np.mean([learning_curves.results_dict[alg_name][ds_name + '_256x16']
                            for ds_name in ds_names], axis=0) for alg_name in alg_names}

    for alg_name in alg_names:
        axs.plot(n_train, alg_results[alg_name] - alg_results['NN_random'], '--o', label=escape(alg_name))

    axs.legend()
    plt.tight_layout()
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name)
    plt.close(fig)


def alt_plot_eff_dim(results: ExperimentResults, filename: Union[str, Path], pattern: str, metric_name: str):
    # plots, for every data set, error of grad_rp-512 to ll vs difference in effective dimensions (average),
    # for different selection methods (esp. maxdist, maxdet, lcmd, fw),
    all_alg_names = [pattern.replace('*', 'll'), pattern.replace('*', 'grad_rp-512'), 'NN_random']
    all_task_names = {task_name for alg_name in all_alg_names for task_name in results.results_dict[alg_name].keys()}

    eff_dims = results.get_avg_al_stats('eff_dim')
    last_errors = results.get_last_errors(metric_name)
    ds_names = list({'_'.join(task_name.split('_')[:-1]) for task_name in all_task_names})

    fig, axs = plt.subplots(figsize=(5, 5))

    for name_task in ds_names:
        log_means_ll = []
        log_means_grad_rp = []
        log_means_random = []
        for key in last_errors.results_dict[all_alg_names[0]]:
            if key.startswith(name_task) and int(key.split('_')[-1].split('x')[0]) == 256:
                log_means_ll = [np.mean(last_errors.results_dict[all_alg_names[0]][key]),
                                np.mean(eff_dims.results_dict[all_alg_names[0]][key])]
                log_means_grad_rp = [np.mean(last_errors.results_dict[all_alg_names[1]][key]),
                                     np.mean(eff_dims.results_dict[all_alg_names[1]][key])]
                log_means_random = [np.mean(last_errors.results_dict[all_alg_names[2]][key]),
                                np.mean(eff_dims.results_dict[all_alg_names[2]][key])]

        log_means_ll = np.asarray(log_means_ll)
        log_means_grad_rp = np.asarray(log_means_grad_rp)
        log_means_random = np.asarray(log_means_random)

        axs.plot(log_means_grad_rp[1] - log_means_ll[1],
                 (log_means_ll[0] - log_means_grad_rp[0]) / (abs(log_means_grad_rp[0] - log_means_random[0])
                                                             + abs(log_means_ll[0] - log_means_random[0])),
                 'o', label=escape(name_task))

    axs.legend()
    plt.tight_layout()
    plot_name = Path(custom_paths.get_plots_path()) / results.exp_name / filename
    utils.ensureDir(plot_name)
    plt.savefig(plot_name)
    plt.close(fig)



import numpy as np
import torch
from typing import List

from .data import DictDataset, ParallelDictDataLoader
from .models import create_tabular_model
from .task_execution import get_devices
from .train import fit_model


class NNRegressor:
    """
    Scikit-learn style interface for the NN regression (without active learning) used in this repository.
    """
    def __init__(self, lr: float = 0.15, hidden_sizes: List[int] = None, act: str = 'silu', n_ensemble: int = 1,
                 batch_size: int = 256, n_epochs: int = 256, weight_decay: float = 0.0,
                 weight_gain: float = 0.5, bias_gain: float = 1.0, valid_fraction: float = 0.1, device: str = None,
                 preprocess_data: bool = True, seed: int = 0):
        """
        Constructor with sensible default values (optimized as in the paper).
        :param lr: Learning rate.
        :param hidden_sizes: Sizes of hidden layers. If None, set to [512, 512]
        :param act: Activation function such as 'relu' or 'silu'. For more options, see layers.get_act_layer().
        :param n_ensemble: How many NNs should be used in the ensemble. Defaults to 1.
        :param batch_size: Batch size to use (will be automatically adjusted to a smaller one if needed).
        :param n_epochs: Number of epochs to train maximally.
        :param weight_decay: Weight decay parameter.
        :param weight_gain: Factor for the weight parameters of linear layers.
        :param bias_gain: Factor for the bias parameters of linear layers.
        :param valid_fraction: Which fraction of the training data set should be used for validation.
        :param device: Device to train on. Should be a string that PyTorch accepts, such as 'cpu' or 'cuda:0'.
        If None, the first GPU is used if one is found, otherwise the CPU is used.
        :param preprocess_data: Whether X and y values should be standardized for training and X should be soft-clipped.
        :param seed: Random seed for training.
        """
        self.n_models = n_ensemble
        self.hidden_sizes = hidden_sizes or [512, 512]
        self.lr = lr
        self.act = act
        self.weight_gain = weight_gain
        self.bias_gain = bias_gain
        self.model = None
        self.valid_fraction = valid_fraction
        self.device = device or get_devices()[0]
        self.seed = seed
        self.batch_size = batch_size
        self.n_epochs = n_epochs
        self.weight_decay = weight_decay
        self.preprocess_data = preprocess_data
        self.means = None
        self.stds = None
        self.y_mean = None
        self.y_std = None

    def fit(self, X: np.ndarray, y: np.ndarray):
        n_features = X.shape[1]
        if len(y.shape) == 1:
            y = y[:, None]
        n_outputs = y.shape[1]
        np.random.seed(self.seed)
        torch.manual_seed(self.seed)

        self.model = create_tabular_model(n_models=self.n_models, n_features=n_features,
                                          hidden_sizes=self.hidden_sizes,
                                          act=self.act, n_outputs=n_outputs,
                                          weight_gain=self.weight_gain, bias_gain=self.bias_gain).to(self.device)

        X = torch.as_tensor(X, dtype=torch.float).to(self.device)
        y = torch.as_tensor(y, dtype=torch.float).to(self.device)

        if self.preprocess_data:
            self.means = X.mean(dim=0, keepdim=True)
            self.stds = X.std(dim=0, keepdim=True)
            self.y_mean = y.mean().item()
            self.y_std = y.std().item()
            X = (X - self.means) / (self.stds + 1e-30)
            X = 5 * torch.tanh(0.2 * X)
            y = (y - self.y_mean) / (self.y_std + 1e-30)

        data = DictDataset({'X': X, 'y': y})
        n_valid = int(self.valid_fraction * X.shape[0])
        perm = np.random.permutation(X.shape[0])
        valid_idxs = torch.as_tensor(perm[:n_valid]).to(self.device)
        train_idxs = torch.as_tensor(perm[n_valid:]).to(self.device)

        fit_model(self.model, data, n_models=self.n_models, train_idxs=train_idxs, valid_idxs=valid_idxs,
                  n_epochs=self.n_epochs, batch_size=self.batch_size, lr=self.lr,
                  weight_decay=self.weight_decay, valid_batch_size=8192)

    def predict(self, X):
        X = torch.as_tensor(X, dtype=torch.float).to(self.device)

        if self.preprocess_data:
            X = (X - self.means) / (self.stds + 1e-30)
            X = 5 * torch.tanh(0.2 * X)

        data = DictDataset({'X': X})
        idxs = torch.arange(X.shape[0], device=self.device)

        dl = ParallelDictDataLoader(data, idxs.expand(self.n_models, -1), batch_size=8192, shuffle=False,
                                         adjust_bs=False, drop_last=False)
        with torch.no_grad():
            self.model.eval()
            y_pred = torch.cat([self.model(batch['X']) for batch in dl], dim=1).mean(dim=0)

        if self.preprocess_data:
            y_pred = y_pred * self.y_std + self.y_mean

        return y_pred.detach().cpu().numpy()







from abc import ABC
import numpy as np

import torch
from typing import *

from .. import utils
from .features import *


class SelectionMethod:
    """
    Abstract base class for selection methods,
    which allow to select a subset of indices from the pool set as the next batch to label for Batch Active Learning.
    """
    def __init__(self):
        super().__init__()
        self.status = None  # can be used to report errors during selection

    def select(self, batch_size: int) -> torch.Tensor:
        """
        Select batch_size elements from the pool set
        (which is assumed to be given in the constructor of the corresponding subclass).
        This method needs to be implemented by subclasses.
        It is assumed that this method is only called once per object, since it can modify the state of the object.
        :param batch_size: Number of elements to select from the pool set.
        :return: Returns a torch.Tensor of integer type that contains the selected indices.
        """
        raise NotImplementedError()

    def get_status(self) -> Optional:
        """
        :return: Returns an object representing the status of the selection. If all went well, the method returns None.
        Otherwise it might return a string or something different representing an error that occured.
        This is mainly useful for analyzing a lot of experiment runs.
        """
        return self.status


class IterativeSelectionMethod(SelectionMethod):
    """
    Abstract base class for iterative selection methods, as considered in the paper.
    """
    def __init__(self, pool_features: Features, train_features: Features, sel_with_train: bool, verbosity: int = 1,
                 **config):
        """
        :param pool_features: Features representing the pool set.
        :param train_features: Features representing the training set.
        :param sel_with_train: This corresponds to the mode parameter in the paper.
        Set to True if you want to use TP-mode (i.e. use the training data for selection), and to False for P-mode.
        :param verbosity: Level of verbosity. If >= 1, something may be printed to indicate the progress of selection.
        """
        super().__init__()
        self.train_features = train_features
        self.pool_features = pool_features
        self.features = pool_features.concat_with(train_features) if sel_with_train else pool_features
        self.selected_idxs = []
        self.selected_arr = torch.zeros(self.pool_features.get_n_samples(), dtype=torch.bool,
                                        device=self.pool_features.get_device())
        self.with_train = sel_with_train
        self.verbosity = verbosity

    def prepare(self, n_adds: int):
        """
        Callback method that may be implemented by subclasses.
        This method is called before starting the selection.
        It can be used, for example, to allocate memory depending on the batch size.
        :param n_adds: How often add() will be called during select()
        """
        pass

    def get_scores(self) -> torch.Tensor:
        """
        Abstract method that can be implemented by subclasses.
        This method should return a score for each pool sample,
        which can then be used to select the next pool sample to include.
        :return: Returns a torch.Tensor of shape [len(self.pool_features)] containing the scores.
        """
        raise NotImplementedError()

    def add(self, new_idx: int):
        """
        Update the state of the object (and therefore the scores) based on adding a new point to the selected set.
        :param new_idx: idx of the new point wrt to self.features,
        i.e. if new_idx > len(self.pool_features),
        then new_idx - len(self.pool_features) is the index for self.train_features,
        otherwise new_idx is an index to self.pool_features.
        """
        raise NotImplementedError()

    def get_next_idx(self) -> Optional[int]:
        """
        This method may be overridden by subclasses.
        It should return the index of the next sample that should be added to the batch.
        By default, it returns the index with the maximum score, according to self.get_scores().
        :return: Returns an int corresponding to the next index.
        It may also return None to indicate that the selection of the next index failed
        and that the batch should be filled up with random samples.
        """
        scores = self.get_scores().clone()
        scores[self.selected_idxs] = -np.Inf
        return torch.argmax(self.get_scores()).item()

    def select(self, batch_size: int) -> torch.Tensor:
        """
        Iterative implementation of batch selection for Batch Active Learning.
        :param batch_size: Number of elements that should be included in the batch.
        :return: Returns a torch.Tensor of integer type containing the indices of the selected pool set elements.
        """
        device = self.pool_features.get_device()

        self.prepare(batch_size + len(self.train_features) if self.with_train else batch_size)

        if self.with_train:
            # add training points first
            for i in range(len(self.train_features)):
                self.add(len(self.pool_features)+i)
                if (i+1) % 256 == 0 and self.verbosity >= 1:
                    print(f'Added {i+1} train samples to selection', flush=True)

        for i in range(batch_size):
            next_idx = self.get_next_idx()
            if next_idx is None or next_idx < 0 or next_idx >= len(self.pool_features) or self.selected_arr[next_idx]:
                # data selection failed
                # fill up with random remaining indices
                self.status = f'filling up with random samples because selection failed after n_selected = {len(self.selected_idxs)}'
                if self.verbosity >= 1:
                    print(self.status)
                n_missing = batch_size - len(self.selected_idxs)
                remaining_idxs = torch.nonzero(~self.selected_arr).squeeze(-1)
                new_random_idxs = remaining_idxs[torch.randperm(len(remaining_idxs), device=device)[:n_missing]]
                selected_idxs_tensor = torch.as_tensor(self.selected_idxs, dtype=torch.long,
                                                       device=torch.device(device))
                return torch.cat([selected_idxs_tensor, new_random_idxs], dim=0)
            else:
                self.add(next_idx)
                self.selected_idxs.append(next_idx)
                self.selected_arr[next_idx] = True
        return torch.as_tensor(self.selected_idxs, dtype=torch.long, device=torch.device(device))


class ForwardBackwardSelectionMethod(IterativeSelectionMethod, ABC):
    """
    This class represents a forward-backward selection method that first selects too many samples
    and then removes some of them. It is an extension of the iterative selection method template.
    This is used for the forward-backward version of the BAIT selection method.
    """
    def __init__(self, pool_features: Features, train_features: Features, sel_with_train: bool, verbosity: int = 1,
                 overselection_factor: float = 1.0, **config):
        super().__init__(pool_features=pool_features, train_features=train_features, sel_with_train=sel_with_train,
                         verbosity=verbosity, **config)
        self.overselection_factor = overselection_factor

    def get_scores_backward(self) -> torch.Tensor:
        """
        Abstract method that can be implemented by subclasses.
        This method should return a score for each selected sample,
        which can then be used to select the next pool sample to include.
        :return: Returns a torch.Tensor of shape [len(self.selected_idxs)] containing the scores.
        """
        raise NotImplementedError()

    def get_next_idx_backward(self) -> Optional[int]:
        """
        This method may be overridden by subclasses.
        It should return the index of the next sample that should be removed from the batch.
        By default, it returns the index with the maximum score, according to self.get_scores_backward().
        :return: Returns an int corresponding to the next index to be removed.
        It may also return None to indicate that the selection of the next index failed
        and that the batch should be filled up with random samples.
        """
        scores = self.get_scores_backward().clone()
        return torch.argmax(scores).item()

    def remove(self, idx: int):
        """
        Update the state of the object (and therefore the scores) based on removing a point from the selected set.
        :param idx: index of the new point wrt to self.selected_idxs.
        """
        raise NotImplementedError()

    def select(self, batch_size: int) -> torch.Tensor:
        """
        Iterative implementation of batch selection for Batch Active Learning.
        :param batch_size: Number of elements that should be included in the batch.
        :return: Returns a torch.Tensor of integer type containing the indices of the selected pool set elements.
        """
        overselect_batch_size = min(round(self.overselection_factor * batch_size), len(self.pool_features))
        overselect_batch = super().select(overselect_batch_size)
        if batch_size == overselect_batch_size:
            return overselect_batch
        if self.status is not None:
            # selection failed
            return overselect_batch[:batch_size]

        for i in range(overselect_batch_size-batch_size):
            next_idx = self.get_next_idx_backward()
            if next_idx is None or next_idx < 0 or next_idx >= len(self.selected_idxs):
                self.status = f'removing the latest overselected samples because the backward step failed '\
                              f'after removing {i} samples'
                if self.verbosity >= 1:
                    print(self.status)
                self.selected_idxs = self.selected_idxs[:batch_size]
                break
            else:
                self.remove(next_idx)
                self.selected_arr[self.selected_idxs[next_idx]] = False
                del self.selected_idxs[next_idx]
        device = self.pool_features.get_device()
        return torch.as_tensor(self.selected_idxs, dtype=torch.long, device=torch.device(device))


class RandomSelectionMethod(SelectionMethod):
    """
    Implements random selection for Batch Active Learning.
    """
    def __init__(self, pool_features: Features, **config):
        super().__init__()
        self.pool_features = pool_features

    def select(self, batch_size: int) -> torch.Tensor:
        device = self.pool_features.get_device()
        generator = torch.Generator(device=device)
        return torch.randperm(self.pool_features.get_n_samples(),
                              device=self.pool_features.get_device(),
                              generator=generator)[:batch_size]


class MaxDiagSelectionMethod(SelectionMethod):
    """
    Implements MaxDiag, i.e. naive active learning, for Batch Active Learning.
    """
    def __init__(self, pool_features: Features, **config):
        """
        :param pool_features: Features on the pool set.
        """
        super().__init__()
        self.pool_features = pool_features

    def select(self, batch_size: int) -> torch.Tensor:
        diag = self.pool_features.get_kernel_matrix_diag()
        return torch.argsort(diag)[-batch_size:]


class MaxDetSelectionMethod(IterativeSelectionMethod):
    """
    Implements the kernel-space version of MaxDet for Batch Active Learning.
    """
    def __init__(self, pool_features: Features, train_features: Features, sel_with_train: bool = False,
                 noise_sigma: float = 0.0, **config):
        """
        :param pool_features: Features on the pool set.
        :param train_features: Features on the train set.
        :param sel_with_train: If True, TP-mode is used instead of P-mode. This is False by default.
        :param noise_sigma: noise_sigma**2 is added to the kernel diagonal for determinant maximization.
        """
        super().__init__(pool_features=pool_features, train_features=train_features,
                         sel_with_train=sel_with_train, **config)
        self.noise_sigma = noise_sigma
        self.diag = self.features.get_kernel_matrix_diag() + self.noise_sigma**2
        self.l = None
        self.n_added = 0

    def prepare(self, n_adds: int):
        #
        # usually self.l is None, since select() should only be called once
        n_total = n_adds if self.l is None else n_adds + self.l.shape[1]
        new_l = torch.zeros(len(self.features), n_total, device=torch.device(self.features.get_device()), dtype=self.diag.dtype)
        if self.l is not None:
            new_l[:, :self.l.shape[1]] = self.l
        self.l = new_l

    def get_scores(self) -> torch.Tensor:
        return self.diag

    def get_next_idx(self) -> Optional[int]:
        # print('max score:', torch.max(self.get_scores()).item())
        scores = self.get_scores().clone()
        scores[self.selected_idxs] = -np.Inf
        new_idx = torch.argmax(scores).item()
        if scores[new_idx] <= 0.0:
            print(f'Selecting index {len(self.selected_idxs)+1}: new diag entry nonpositive')
            # print(f'diag entry: {self.get_scores()[new_idx]}')
            # diagonal is zero or lower, would cause numerical errors afterwards
            return None
        # print(self.diag[new_idx])
        return new_idx

    def add(self, new_idx: int):
        # print('new_idx:', new_idx)
        l = None if self.l is None else self.l[:, :self.n_added]
        lTl = 0.0 if l is None else l.matmul(l[new_idx, :])
        mat_col = self.features[new_idx].get_kernel_matrix(self.features).squeeze(0)
        if self.noise_sigma > 0.0:
            mat_col[new_idx] += self.noise_sigma**2
        update = (1.0 / torch.sqrt(self.diag[new_idx])) * (mat_col - lTl)
        # shape: len(self.features)
        self.diag -= update ** 2
        # shape: (n-1) x len(self.features)
        self.l[:, self.n_added] = update
        # self.l = update[:, None] if self.l is None else torch.cat([self.l, update[:, None]], dim=1)
        # print('trace(ll^T):', (self.l**2).sum())

        self.n_added += 1

        # if str(self.pool_features.get_device()) != 'cpu':
        #     torch.cuda.empty_cache()

        self.diag[new_idx] = -np.Inf   # ensure that the index is not selected again


class MaxDetFeatureSpaceSelectionMethod(IterativeSelectionMethod):
    """
    Implements MaxDet in feature space for Batch Active Learning.
    In terms of efficiency, this is only recommended for (roughly) batch_size > 3*n_features,
    so normally MaxDetSelectionMethod should be used instead.
    """
    def __init__(self, pool_features: Features, train_features: Features, sel_with_train: bool = False,
                 noise_sigma: float = 0.0, **config):
        super().__init__(pool_features=pool_features, train_features=train_features,
                         sel_with_train=sel_with_train, **config)
        self.noise_sigma = noise_sigma
        self.diag = self.features.get_kernel_matrix_diag().clone() + self.noise_sigma**2
        self.feature_matrix = self.features.get_feature_matrix().clone()

    def get_scores(self) -> torch.Tensor:
        return self.diag

    def get_next_idx(self) -> Optional[int]:
        scores = self.get_scores().clone()
        scores[self.selected_idxs] = -np.Inf
        new_idx = torch.argmax(scores).item()
        if scores[new_idx] <= 0.0:
            if self.verbosity >= 1:
                print(f'Selecting index {len(self.selected_idxs)+1}: new diag entry nonpositive')
            # diagonal is zero or lower, would cause numerical errors afterwards
            return None
        return new_idx

    def add(self, new_idx: int):
        diag_entry = self.diag[new_idx]
        sqrt_diag_entry = torch.sqrt(diag_entry)
        beta = 1.0 / (sqrt_diag_entry * (sqrt_diag_entry + self.noise_sigma))
        phi_x = self.feature_matrix[new_idx]
        dot_prods = self.feature_matrix.matmul(phi_x)
        self.feature_matrix -= dot_prods[:, None] * (beta * phi_x[None, :])
        self.diag -= dot_prods**2 / diag_entry
        
        
class BaitFeatureSpaceSelectionMethod(ForwardBackwardSelectionMethod):
    """
    Implements BAIT in feature space for Batch Active Learning.
    """
    def __init__(self, pool_features: Features, train_features: Features, sel_with_train: bool = False,
                 noise_sigma: float = 0.0, **config):
        super().__init__(pool_features=pool_features, train_features=train_features,
                         sel_with_train=sel_with_train, **config)
        self.noise_sigma = noise_sigma
        self.diag = self.features.get_kernel_matrix_diag().clone()
        self.feature_matrix = self.features.get_feature_matrix().clone()

        self.feature_cov_matrix = self.feature_matrix.t() @ self.feature_matrix
        if not sel_with_train:
            train_feature_matrix = train_features.get_feature_matrix()
            self.feature_cov_matrix += train_feature_matrix.t() @ train_feature_matrix
        self.scores_numerator = torch.einsum('ij,ji->i',
                                             self.feature_matrix,
                                             self.feature_cov_matrix @ self.feature_matrix.t())

    def get_scores(self) -> torch.Tensor:
        return self.scores_numerator / (self.diag + self.noise_sigma**2 + 1e-8)

    def get_next_idx(self) -> Optional[int]:
        scores = self.get_scores()
        scores[self.selected_idxs] = -np.Inf
        new_idx = torch.argmax(scores).item()
        if scores[new_idx].item() <= 0.0:
            if self.verbosity >= 1:
                print(f'Selecting index {len(self.selected_idxs)+1}: new score nonpositive')
            return None
        return new_idx

    def add(self, new_idx: int):
        diag_entry = self.diag[new_idx] + self.noise_sigma**2
        sqrt_diag_entry = torch.sqrt(diag_entry)
        beta = 1.0 / (sqrt_diag_entry * (sqrt_diag_entry + self.noise_sigma))
        phi_x = self.feature_matrix[new_idx]
        dot_prods = self.feature_matrix.matmul(phi_x)
        dot_prods_sq = dot_prods**2

        # update scores_numerator
        cov_phi = self.feature_cov_matrix @ phi_x
        phi_cov_phi = self.scores_numerator[new_idx].clone()
        # phi_cov_phi = torch.dot(phi_x, cov_phi)
        mult = 1/diag_entry
        self.scores_numerator -= 2 * mult * (self.feature_matrix @ cov_phi) * dot_prods
        self.scores_numerator += mult**2 * phi_cov_phi * dot_prods_sq
        # update feature_cov_matrix
        cov_phi_phit = cov_phi[:, None] * phi_x[None, :]
        phi_phit = phi_x[:, None] * phi_x[None, :]
        self.feature_cov_matrix -= beta * (cov_phi_phit + cov_phi_phit.t())
        self.feature_cov_matrix += beta**2 * phi_cov_phi * phi_phit
        # update feature matrix
        self.feature_matrix -= dot_prods[:, None] * (beta * phi_x[None, :])
        # update diag
        self.diag -= dot_prods_sq / diag_entry


    def get_scores_backward(self) -> torch.Tensor:
        den = (self.diag[self.selected_idxs] - self.noise_sigma ** 2)
        num = torch.clamp(self.scores_numerator[self.selected_idxs], min=0.0)
        scores = num / den
        scores[den >= 0.0] = -np.Inf
        return scores

    def get_next_idx_backward(self) -> Optional[int]:
        scores = self.get_scores_backward()
        new_idx = torch.argmax(scores).item()
        new_score = scores[new_idx].item()
        if new_score == -np.Inf or new_score >= 0.0:
            if self.verbosity >= 1:
                print(f'Backwards selecting index {len(self.selected_idxs)}: new score positive')
            return None
        return new_idx

    def remove(self, idx: int):
        features_idx = self.selected_idxs[idx]
        diag_entry = self.noise_sigma ** 2 - self.diag[features_idx]
        diag_entry = torch.clamp(diag_entry, min=1e-15)
        sqrt_diag_entry = torch.sqrt(diag_entry)
        beta = 1.0 / (sqrt_diag_entry * (sqrt_diag_entry + self.noise_sigma))
        phi_x = self.feature_matrix[features_idx]
        dot_prods = self.feature_matrix.matmul(phi_x)
        dot_prods_sq = dot_prods**2

        # update scores_numerator
        cov_phi = self.feature_cov_matrix @ phi_x
        phi_cov_phi = self.scores_numerator[features_idx].clone()
        mult = 1 / diag_entry
        self.scores_numerator += 2 * mult * (self.feature_matrix @ cov_phi) * dot_prods
        self.scores_numerator += mult ** 2 * phi_cov_phi * dot_prods_sq
        # update feature_cov_matrix
        cov_phi_phit = cov_phi[:, None] * phi_x[None, :]
        phi_phit = phi_x[:, None] * phi_x[None, :]
        self.feature_cov_matrix += beta * (cov_phi_phit + cov_phi_phit.t())
        self.feature_cov_matrix += beta ** 2 * phi_cov_phi * phi_phit
        # update feature matrix
        self.feature_matrix += dot_prods[:, None] * (beta * phi_x[None, :])
        # update diag
        self.diag += dot_prods_sq / diag_entry


class MaxDistSelectionMethod(IterativeSelectionMethod):
    """
    Implements the MaxDist selection method for Batch Active Learning.
    """
    def __init__(self, pool_features: Features, train_features: Features, sel_with_train: bool = True,
                 **config):
        super().__init__(pool_features=pool_features, train_features=train_features,
                         sel_with_train=sel_with_train, **config)
        self.min_sq_dists = np.Inf * torch.ones(self.pool_features.get_n_samples(), device=pool_features.get_device())

    def get_scores(self) -> torch.Tensor:
        return self.min_sq_dists

    def get_next_idx(self) -> Optional[int]:
        if len(self.selected_idxs) == 0:
            # no point added yet, take point with largest norm
            return torch.argmax(self.pool_features.get_kernel_matrix_diag()).item()
        scores = self.get_scores().clone()
        scores[self.selected_idxs] = -np.Inf
        idx = torch.argmax(scores).item()
        # print('Next idx:', idx, '- Value:', scores[idx].item())
        return idx

    def add(self, new_idx: int):
        sq_dists = self.features[new_idx].get_sq_dists(self.pool_features).squeeze(0)
        self.min_sq_dists = torch.minimum(self.min_sq_dists, sq_dists)
        # print('min_sq_dists:', self.min_sq_dists)
        # if new_idx < len(self.pool_features):
        #     print('sq dists at new idx:', sq_dists[new_idx].item(), 'and', self.min_sq_dists[new_idx].item())


class LargestClusterMaxDistSelectionMethod(IterativeSelectionMethod):
    """
    Implements the LCMD selection method for Batch Active Learning.
    """
    def __init__(self, pool_features: Features, train_features: Features, sel_with_train: bool = True,
                 dist_weight_mode: str = 'sq-dist', **config):
        """
        :param pool_features:
        :param train_features:
        :param sel_with_train:
        :param dist_weight_mode: one of 'none', 'dist' or 'sq-dist'
        """
        super().__init__(pool_features=pool_features, train_features=train_features,
                         sel_with_train=sel_with_train, **config)
        self.dist_weight_mode = dist_weight_mode
        self.min_sq_dists = np.Inf * torch.ones(self.pool_features.get_n_samples(), dtype=pool_features.get_dtype(),
                                                device=pool_features.get_device())
        self.closest_idxs = torch.zeros(self.pool_features.get_n_samples(), device=pool_features.get_device(),
                                        dtype=torch.long)
        self.neg_inf_tensor = torch.as_tensor([-np.Inf], dtype=pool_features.get_dtype(),
                                              device=pool_features.get_device())

        self.n_added = 0

    def get_scores(self) -> torch.Tensor:
        if self.dist_weight_mode == 'sq-dist':
            weights = self.min_sq_dists
        elif self.dist_weight_mode == 'dist':
            weights = self.min_sq_dists.sqrt()
        else:
            weights = None
        bincount = torch.bincount(self.closest_idxs, weights=weights, minlength=self.n_added+1)
        max_bincount = torch.max(bincount)
        # print(f'max bincount: {max_bincount.item():g}, max_min_sq_dist: {torch.max(self.min_sq_dists).item():g}, '
        #       f'number of zero dists: {self.min_sq_dists.shape[0] - torch.count_nonzero(self.min_sq_dists).item()}, ',
        #       f'average kernel diag: {self.features.get_kernel_matrix_diag().mean().item():g}')
        return torch.where(bincount[self.closest_idxs] == max_bincount, self.min_sq_dists, self.neg_inf_tensor)

    def get_next_idx(self) -> Optional[int]:
        if len(self.selected_idxs) == 0:
            # no point added yet, take point with largest norm
            return torch.argmax(self.pool_features.get_kernel_matrix_diag()).item()
        scores = self.get_scores().clone()
        scores[self.selected_idxs] = -np.Inf
        idx = torch.argmax(scores).item()
        return idx

    def add(self, new_idx: int):
        sq_dists = self.features[new_idx].get_sq_dists(self.pool_features).squeeze(0)
        self.n_added += 1
        new_min = sq_dists < self.min_sq_dists
        self.closest_idxs[new_min] = self.n_added
        self.min_sq_dists[new_min] = sq_dists[new_min]


class FrankWolfeSelectionMethod(IterativeSelectionMethod):
    """
    Implements the FrankWolfe selection method in feature space for Batch Active Learning.
    """
    def __init__(self, pool_features: Features, train_features: Features, sel_with_train: bool = False,
                 **config):
        super().__init__(pool_features=pool_features, train_features=train_features,
                         sel_with_train=sel_with_train, **config)
        if pool_features.get_n_features() <= 0:
            raise NotImplementedError('Frank-Wolfe is currently only implemented for finite-dimensional feature maps '
                                      'due to quadratic time in n_pool for infinite-dimensional feature maps')
        self.sqrt_diag = self.features.get_kernel_matrix_diag().sqrt()
        self.sqrt_diag_sum = self.sqrt_diag.sum()
        self.feature_matrix = self.features.get_feature_matrix()
        self.eps = 1e-30
        self.normalized_feature_matrix = self.feature_matrix / (self.sqrt_diag[:, None] + self.eps)
        self.kernel_sum_embedding = self.feature_matrix.sum(dim=0)
        self.current_embedding = torch.zeros(self.feature_matrix.shape[-1], device=self.feature_matrix.device)

    def get_scores(self) -> torch.Tensor:
        return self.normalized_feature_matrix.matmul(self.kernel_sum_embedding - self.current_embedding)

    def get_next_idx(self) -> Optional[int]:
        scores = self.get_scores()
        scores[self.selected_idxs] = -np.Inf
        return torch.argmax(scores).item()

    def add(self, new_idx: int):
        normalized_feature_vector = self.normalized_feature_matrix[new_idx, :]
        residual = self.kernel_sum_embedding - self.current_embedding
        current_diff = self.sqrt_diag_sum * normalized_feature_vector - self.current_embedding
        gamma = torch.dot(current_diff, residual) / torch.dot(current_diff, current_diff)
        self.current_embedding = (1 - gamma) * self.current_embedding \
                                 + gamma * self.sqrt_diag_sum * normalized_feature_vector


class FrankWolfeKernelSpaceSelectionMethod(IterativeSelectionMethod):
    """
    Implements the FrankWolfe selection method in kernel space for Batch Active Learning.
    Note that runtime of the kernel-space version here scales quadratically with the pool set size.
    Therefore, it should be avoided for large pool sets. The feature-space version, FrankWolfeSelectionMethod,
    scales better for moderately large feature space dimensions.
    """
    def __init__(self, pool_features: Features, train_features: Features, sel_with_train: bool = False,
                 **config):
        super().__init__(pool_features=pool_features, train_features=train_features,
                         sel_with_train=sel_with_train, **config)
        if pool_features.get_n_features() <= 0:
            raise NotImplementedError('Frank-Wolfe is currently only implemented for finite-dimensional feature maps '
                                      'due to quadratic time in n_pool for infinite-dimensional feature maps')
        self.sqrt_diag = self.features.get_kernel_matrix_diag().sqrt()  # c
        self.sqrt_diag_sum = self.sqrt_diag.sum()  # r
        batch_size = 2**22 // len(self.features) + 1
        batch_intervals = utils.get_batch_intervals(len(self.features), batch_size=batch_size)
        self.u = torch_cat([self.features[start:stop].get_kernel_matrix(self.features).sum(dim=-1)
                            for start, stop in batch_intervals], dim=0)
        self.v = torch.zeros_like(self.u)
        self.s = 0.0
        self.t = 0.0
        self.eps = 1e-30

    def get_scores(self) -> torch.Tensor:
        return (self.u - self.v) / (self.sqrt_diag + self.eps)

    def get_next_idx(self) -> Optional[int]:
        scores = self.get_scores()
        scores[self.selected_idxs] = -np.Inf
        return torch.argmax(scores).item()

    def add(self, new_idx: int):
        rcinv = self.sqrt_diag_sum / self.sqrt_diag[new_idx]
        ui = self.u[new_idx]
        vi = self.v[new_idx]
        denominator = (self.sqrt_diag_sum**2 - 2*rcinv*vi + self.s)
        denominator = max(denominator, self.eps)
        gamma = (rcinv * (ui - vi) + self.s - self.t) / denominator
        self.s = (1-gamma)**2 * self.s + 2*(1-gamma)*gamma*rcinv*vi + gamma**2 * self.sqrt_diag_sum**2
        self.t = (1-gamma)*self.t + gamma * rcinv * ui
        self.v = (1-gamma)*self.v + gamma * rcinv * self.features.get_kernel_matrix(self.features[new_idx])[:, 0]


class KmeansppSelectionMethod(MaxDistSelectionMethod):
    """
    Implements the KMeansPP (k-means++) selection method for Batch Active Learning.
    This class inherits from MaxDistSelectionMethod to reuse the squared-distance computations.
    """
    def __init__(self, pool_features: Features, train_features: Features, sel_with_train: bool = True,
                 **config):
        super().__init__(pool_features=pool_features, train_features=train_features,
                         sel_with_train=sel_with_train, **config)

    def get_next_idx(self) -> Optional[int]:
        if len(self.selected_idxs) == 0:
            return np.random.randint(low=0, high=len(self.pool_features))
        try:
            return torch.multinomial(torch.clamp(self.min_sq_dists, min=0.0), 1).item()
        except RuntimeError:
            return None  # failed, potentially because the sum of the dists was <= 0


class RandomizedMinDistSumSelectionMethod(MaxDistSelectionMethod):
    """
    Experimental selection method.
    """
    def __init__(self, pool_features: Features, train_features: Features, sel_with_train: bool = True,
                 max_n_candidates: int = 5, **config):
        super().__init__(pool_features=pool_features, train_features=train_features,
                         sel_with_train=sel_with_train, **config)
        self.max_n_candidates = max_n_candidates

    def get_next_idx(self) -> Optional[int]:
        n_candidates = min(self.max_n_candidates, len(self.pool_features) - len(self.selected_idxs))
        weights = torch.ones_like(self.min_sq_dists) if len(self.selected_idxs) == 0 \
            else torch.clamp(self.min_sq_dists, min=0.0)
        candidates = torch.multinomial(weights, n_candidates)
        sq_dists = self.pool_features[candidates].get_sq_dists(self.pool_features)
        new_sq_dist_sums = torch.minimum(self.min_sq_dists[None, :].expand(n_candidates, -1), sq_dists).sum(dim=-1)
        new_sq_dist_sums[self.selected_idxs] = np.Inf
        return candidates[torch.argmin(new_sq_dist_sums)].item()


class SumOfSquaredDistsSelectionMethod(MaxDistSelectionMethod):
    """
    Experimental selection method. Scales quadratically with the pool set size.
    """
    def get_scores(self) -> torch.Tensor:
        print('computing scores')
        comp_batch_size = 512
        batch_intervals = utils.get_batch_intervals(len(self.pool_features), batch_size=comp_batch_size)
        return -torch_cat([
            torch.minimum(self.min_sq_dists[None, :].expand(stop-start, -1),
                          self.pool_features[start:stop].get_sq_dists(self.pool_features)).sum(dim=1)
            for start, stop in batch_intervals
        ], dim=0)


